import matplotlib as mpl

import yt

# Load the dataset.
ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# Create projections using each colormap available.
p = yt.ProjectionPlot(ds, "z", "density", weight_field="density", width=0.4)

for cmap in mpl.colormaps:
    if cmap.startswith("idl"):
        continue
    p.set_cmap(field="density", cmap=cmap)
    p.annotate_title(cmap)
    p.save(f"Projection_{cmap.replace(' ', '_')}.png")



from .physical_ratios import (
    au_per_mpc,
    cm_per_mpc,
    km_per_mpc,
    kpc_per_mpc,
    miles_per_mpc,
    mpc_per_mpc,
    pc_per_mpc,
    rsun_per_mpc,
    sec_per_day,
    sec_per_Gyr,
    sec_per_Myr,
    sec_per_year,
)

# The number of levels we expect to have at most
MAXLEVEL = 48

# How many of each thing are in an Mpc
mpc_conversion = {
    "Mpc": mpc_per_mpc,
    "mpc": mpc_per_mpc,
    "kpc": kpc_per_mpc,
    "pc": pc_per_mpc,
    "au": au_per_mpc,
    "rsun": rsun_per_mpc,
    "miles": miles_per_mpc,
    "km": km_per_mpc,
    "cm": cm_per_mpc,
}

# Nicely formatted versions of common length units
formatted_length_unit_names = {
    "au": "AU",
    "rsun": r"R_\odot",
    "code_length": r"code\ length",
}

# How many seconds are in each thing
sec_conversion = {
    "Gyr": sec_per_Gyr,
    "Myr": sec_per_Myr,
    "years": sec_per_year,
    "days": sec_per_day,
}



# distutils: sources = yt/utilities/lib/origami_tags.c
# distutils: include_dirs = LIB_DIR
# distutils: depends = yt/utilities/lib/origami_tags.h
# distutils: libraries = STD_LIBS
"""
This calls the ORIGAMI routines



"""


import numpy as np

cimport numpy as np
from libc.stdlib cimport free, malloc


cdef extern from "origami_tags.h":
    int compute_tags(int ng, double boxsize, double **r, int npart,
                     unsigned char *m)

cdef int printed_citation = 0

def run_origami(np.ndarray[np.float64_t, ndim=1] pos_x,
                np.ndarray[np.float64_t, ndim=1] pos_y,
                np.ndarray[np.float64_t, ndim=1] pos_z,
                double boxsize):
    # We assume these have been passed in in the correct order and
    # C-contiguous.
    global printed_citation
    if printed_citation == 0:
        print("ORIGAMI was developed by Bridget Falck and Mark Neyrinck.")
        print("Please cite Falck, Neyrinck, & Szalay 2012, ApJ, 754, 2, 125.")
        printed_citation = 1
    cdef int npart = pos_x.size
    if npart == 1:
        return np.zeros(1, dtype="uint8")
    assert(sizeof(unsigned char) == sizeof(np.uint8_t))
    assert(sizeof(double) == sizeof(np.float64_t))
    cdef int ng = np.round(npart**(1./3))
    assert(ng**3 == npart)
    cdef double **r = <double **> malloc(sizeof(double *) * 3)
    r[0] = <double *> pos_x.data
    r[1] = <double *> pos_y.data
    r[2] = <double *> pos_z.data
    cdef np.ndarray[np.uint8_t, ndim=1] tags = np.zeros(npart, dtype="uint8")
    cdef void *m = <void*> tags.data
    compute_tags(ng, boxsize, r, npart, <unsigned char*> m)
    free(r)
    return tags



import yt

ds = yt.load("enzo_tiny_cosmology/DD0046/DD0046")
p = yt.ProjectionPlot(ds, "z", ("gas", "density"))
p.annotate_sphere([0.54, 0.72], radius=(1, "Mpc"), coord_system="axis", text="Halo #7")
p.annotate_sphere(
    [0.65, 0.38, 0.3],
    radius=(1.5, "Mpc"),
    coord_system="data",
    circle_args={"color": "green", "linewidth": 4, "linestyle": "dashed"},
)
p.annotate_arrow([0.87, 0.59, 0.2], coord_system="data", color="red")
p.annotate_text([10, 20], "Some halos", coord_system="plot")
p.annotate_marker([0.45, 0.1, 0.4], coord_system="data", color="yellow", s=500)
p.annotate_line([0.2, 0.4], [0.3, 0.9], coord_system="axis")
p.annotate_timestamp(redshift=True)
p.annotate_scale()
p.save()



from . import sample_data, tests
from .data_structures import (
    StreamDataset,
    StreamGrid,
    StreamHandler,
    StreamHierarchy,
    hexahedral_connectivity,
)
from .fields import StreamFieldInfo
from .io import IOHandlerStream



import numpy as np

from yt.data_objects.construction_data_containers import YTStreamline
from yt.funcs import get_pbar
from yt.units.yt_array import YTArray
from yt.utilities.amr_kdtree.api import AMRKDTree
from yt.utilities.parallel_tools.parallel_analysis_interface import (
    ParallelAnalysisInterface,
    parallel_passthrough,
)


def sanitize_length(length, ds):
    # Ensure that lengths passed in with units are returned as code_length
    # magnitudes without units
    if isinstance(length, YTArray):
        return ds.arr(length).in_units("code_length").d
    else:
        return length


class Streamlines(ParallelAnalysisInterface):
    r"""A collection of streamlines that flow through the volume

    The Streamlines object contains a collection of streamlines
    defined as paths that are parallel to a specified vector field.

    Parameters
    ----------
    ds : ~yt.data_objects.static_output.Dataset
        This is the dataset to streamline
    positions : array_like
        An array of initial starting positions of the streamlines.
    xfield : str or tuple of str, optional
        The x component of the vector field to be streamlined.
        Default:'velocity_x'
    yfield : str or tuple of str, optional
        The y component of the vector field to be streamlined.
        Default:'velocity_y'
    zfield : str or tuple of str, optional
        The z component of the vector field to be streamlined.
        Default:'velocity_z'
    volume : `yt.extensions.volume_rendering.AMRKDTree`, optional
        The volume to be streamlined.  Can be specified for
        finer-grained control, but otherwise will be automatically
        generated.  At this point it must use the AMRKDTree.
        Default: None
    dx : float, optional
        Optionally specify the step size during the integration.
        Default: minimum dx
    length : float, optional
        Optionally specify the length of integration.
        Default: np.max(self.ds.domain_right_edge-self.ds.domain_left_edge)
    direction : real, optional
        Specifies the direction of integration.  The magnitude of this
        value has no effect, only the sign.
    get_magnitude: bool, optional
        Specifies whether the Streamlines.magnitudes array should be
        filled with the magnitude of the vector field at each point in
        the streamline.  This seems to be a ~10% hit to performance.
        Default: False

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    ... import numpy as np
    ... from mpl_toolkits.mplot3d import Axes3D
    ... import yt
    ... from yt.visualization.api import Streamlines

    >>> # Load the dataset and set some parameters
    >>> ds = load("IsolatedGalaxy/galaxy0030/galaxy0030")
    >>> c = np.array([0.5] * 3)
    >>> N = 100
    >>> scale = 1.0
    >>> pos_dx = np.random.random((N, 3)) * scale - scale / 2.0
    >>> pos = c + pos_dx

    >>> # Define and construct streamlines
    >>> streamlines = Streamlines(
    ...     ds, pos, "velocity_x", "velocity_y", "velocity_z", length=1.0
    ... )
    >>> streamlines.integrate_through_volume()

    >>> # Make a 3D plot of the streamlines and save it to disk
    >>> fig = plt.figure()
    >>> ax = Axes3D(fig)
    >>> for stream in streamlines.streamlines:
    ...     stream = stream[np.all(stream != 0.0, axis=1)]
    ...     ax.plot3D(stream[:, 0], stream[:, 1], stream[:, 2], alpha=0.1)
    >>> plt.savefig("streamlines.png")
    """

    def __init__(
        self,
        ds,
        positions,
        xfield="velocity_x",
        yfield="velocity_x",
        zfield="velocity_x",
        volume=None,
        dx=None,
        length=None,
        direction=1,
        get_magnitude=False,
    ):
        ParallelAnalysisInterface.__init__(self)
        self.ds = ds
        self.start_positions = sanitize_length(positions, ds)
        self.N = self.start_positions.shape[0]
        # I need a data object to resolve the field names to field tuples
        # via _determine_fields()
        ad = self.ds.all_data()
        self.xfield = ad._determine_fields(xfield)[0]
        self.yfield = ad._determine_fields(yfield)[0]
        self.zfield = ad._determine_fields(zfield)[0]
        self.get_magnitude = get_magnitude
        self.direction = np.sign(direction)
        if volume is None:
            volume = AMRKDTree(self.ds)
            volume.set_fields(
                [self.xfield, self.yfield, self.zfield], [False, False, False], False
            )
            volume.join_parallel_trees()
        self.volume = volume
        if dx is None:
            dx = self.ds.index.get_smallest_dx()
        self.dx = sanitize_length(dx, ds)
        if length is None:
            length = np.max(self.ds.domain_right_edge - self.ds.domain_left_edge)
        self.length = sanitize_length(length, ds)
        self.steps = int(self.length / self.dx) + 1
        # Fix up the dx.
        self.dx = 1.0 * self.length / self.steps
        self.streamlines = np.zeros((self.N, self.steps, 3), dtype="float64")
        self.magnitudes = None
        if self.get_magnitude:
            self.magnitudes = np.zeros((self.N, self.steps), dtype="float64")

    def integrate_through_volume(self):
        nprocs = self.comm.size
        my_rank = self.comm.rank
        self.streamlines[my_rank::nprocs, 0, :] = self.start_positions[my_rank::nprocs]

        pbar = get_pbar("Streamlining", self.N)
        for i, stream in enumerate(self.streamlines[my_rank::nprocs]):
            thismag = None
            if self.get_magnitude:
                thismag = self.magnitudes[i, :]
            step = self.steps
            while step > 1:
                this_node = self.volume.locate_node(stream[-step, :])
                step = self._integrate_through_brick(
                    this_node, stream, step, mag=thismag
                )
            pbar.update(i + 1)
        pbar.finish()

        self._finalize_parallel(None)
        self.streamlines = self.ds.arr(self.streamlines, "code_length")
        if self.get_magnitude:
            self.magnitudes = self.ds.arr(
                self.magnitudes, self.ds.field_info[self.xfield].units
            )

    @parallel_passthrough
    def _finalize_parallel(self, data):
        self.streamlines = self.comm.mpi_allreduce(self.streamlines, op="sum")
        if self.get_magnitude:
            self.magnitudes = self.comm.mpi_allreduce(self.magnitudes, op="sum")

    def _integrate_through_brick(self, node, stream, step, periodic=False, mag=None):
        LE = self.ds.domain_left_edge.d
        RE = self.ds.domain_right_edge.d
        while step > 1:
            self.volume.get_brick_data(node)
            brick = node.data
            stream[-step + 1] = stream[-step]
            if mag is None:
                brick.integrate_streamline(
                    stream[-step + 1], self.direction * self.dx, None
                )
            else:
                marr = [mag]
                brick.integrate_streamline(
                    stream[-step + 1], self.direction * self.dx, marr
                )
                mag[-step + 1] = marr[0]

            cur_stream = stream[-step + 1, :]
            if np.sum(np.logical_or(cur_stream < LE, cur_stream >= RE)):
                return 0

            nLE = node.get_left_edge()
            nRE = node.get_right_edge()
            if np.sum(np.logical_or(cur_stream < nLE, cur_stream >= nRE)):
                return step - 1
            step -= 1
        return step

    def clean_streamlines(self):
        temp = np.empty(self.N, dtype="object")
        temp2 = np.empty(self.N, dtype="object")
        for i, stream in enumerate(self.streamlines):
            mask = np.all(stream != 0.0, axis=1)
            temp[i] = stream[mask]
            temp2[i] = self.magnitudes[i, mask]
        self.streamlines = temp
        self.magnitudes = temp2

    def path(self, streamline_id):
        """
        Returns an YTSelectionContainer1D object defined by a streamline.

        Parameters
        ----------
        streamline_id : int
            This defines which streamline from the Streamlines object
            that will define the YTSelectionContainer1D object.

        Returns
        -------
        An YTStreamline YTSelectionContainer1D object

        Examples
        --------

        >>> from yt.visualization.api import Streamlines
        >>> streamlines = Streamlines(ds, [0.5] * 3)
        >>> streamlines.integrate_through_volume()
        >>> stream = streamlines.path(0)
        >>> fig, ax = plt.subplots()
        >>> ax.set_yscale("log")
        >>> ax.plot(stream["t"], stream[("gas", "density")], "-x")

        """
        return YTStreamline(
            self.streamlines[streamline_id], ds=self.ds, length=self.length
        )



from abc import abstractmethod

from .parallel_analysis_interface import ProcessorPool


class WorkSplitter:
    def __init__(self, controller, group1, group2):
        self.group1 = group1
        self.group2 = group2
        self.controller = controller

    @classmethod
    def setup(cls, ng1, ng2):
        pp, wg = ProcessorPool.from_sizes(
            [(1, "controller"), (ng1, "group1"), (ng2, "group2")]
        )
        groupc = pp["controller"]
        group1 = pp["group1"]
        group2 = pp["group2"]
        obj = cls(groupc, group1, group2)
        obj.run(wg.name)

    def run(self, name):
        if name == "controller":
            self.run_controller()
        elif name == "group1":
            self.run_group1()
        elif name == "group2":
            self.run_group2()
        else:
            raise NotImplementedError

    @abstractmethod
    def run_controller(self):
        pass

    @abstractmethod
    def run_group1(self):
        pass

    @abstractmethod
    def run_group2(self):
        pass



import matplotlib.pyplot as plt
import numpy as np

import yt

"""
Make a turbulent KE power spectrum.  Since we are stratified, we use
a rho**(1/3) scaling to the velocity to get something that would
look Kolmogorov (if the turbulence were fully developed).

Ultimately, we aim to compute:

                      1  ^      ^*
     E(k) = integral  -  V(k) . V(k) dS
                      2

             n                                               ^
where V = rho  U is the density-weighted velocity field, and V is the
FFT of V.

(Note: sometimes we normalize by 1/volume to get a spectral
energy density spectrum).


"""


def doit(ds):

    # a FFT operates on uniformly gridded data.  We'll use the yt
    # covering grid for this.

    max_level = ds.index.max_level

    ref = int(np.product(ds.ref_factors[0:max_level]))

    low = ds.domain_left_edge
    dims = ds.domain_dimensions * ref

    nx, ny, nz = dims

    nindex_rho = 1.0 / 3.0

    Kk = np.zeros((nx // 2 + 1, ny // 2 + 1, nz // 2 + 1))

    for vel in [("gas", "velocity_x"), ("gas", "velocity_y"), ("gas", "velocity_z")]:

        Kk += 0.5 * fft_comp(
            ds, ("gas", "density"), vel, nindex_rho, max_level, low, dims
        )

    # wavenumbers
    L = (ds.domain_right_edge - ds.domain_left_edge).d

    kx = np.fft.rfftfreq(nx) * nx / L[0]
    ky = np.fft.rfftfreq(ny) * ny / L[1]
    kz = np.fft.rfftfreq(nz) * nz / L[2]

    # physical limits to the wavenumbers
    kmin = np.min(1.0 / L)
    kmax = np.min(0.5 * dims / L)

    kbins = np.arange(kmin, kmax, kmin)
    N = len(kbins)

    # bin the Fourier KE into radial kbins
    kx3d, ky3d, kz3d = np.meshgrid(kx, ky, kz, indexing="ij")
    k = np.sqrt(kx3d**2 + ky3d**2 + kz3d**2)

    whichbin = np.digitize(k.flat, kbins)
    ncount = np.bincount(whichbin)

    E_spectrum = np.zeros(len(ncount) - 1)

    for n in range(1, len(ncount)):
        E_spectrum[n - 1] = np.sum(Kk.flat[whichbin == n])

    k = 0.5 * (kbins[0 : N - 1] + kbins[1:N])
    E_spectrum = E_spectrum[1:N]

    index = np.argmax(E_spectrum)
    kmax = k[index]
    Emax = E_spectrum[index]

    plt.loglog(k, E_spectrum)
    plt.loglog(k, Emax * (k / kmax) ** (-5.0 / 3.0), ls=":", color="0.5")

    plt.xlabel(r"$k$")
    plt.ylabel(r"$E(k)dk$")

    plt.savefig("spectrum.png")


def fft_comp(ds, irho, iu, nindex_rho, level, low, delta):

    cube = ds.covering_grid(level, left_edge=low, dims=delta, fields=[irho, iu])

    rho = cube[irho].d
    u = cube[iu].d

    nx, ny, nz = rho.shape

    # do the FFTs -- note that since our data is real, there will be
    # too much information here.  fftn puts the positive freq terms in
    # the first half of the axes -- that's what we keep.  Our
    # normalization has an '8' to account for this clipping to one
    # octant.
    ru = np.fft.fftn(rho**nindex_rho * u)[
        0 : nx // 2 + 1, 0 : ny // 2 + 1, 0 : nz // 2 + 1
    ]
    ru = 8.0 * ru / (nx * ny * nz)

    return np.abs(ru) ** 2


ds = yt.load("maestro_xrb_lores_23437")
doit(ds)



import os
import tempfile
from collections import OrderedDict

from yt.frontends.arepo.api import ArepoHDF5Dataset
from yt.testing import ParticleSelectionComparison, assert_allclose_units, requires_file
from yt.utilities.answer_testing.framework import data_dir_load, requires_ds, sph_answer

bullet_h5 = "ArepoBullet/snapshot_150.hdf5"
tng59_h5 = "TNGHalo/halo_59.hdf5"
_tng59_bbox = [[40669.34, 56669.34], [45984.04, 61984.04], [54114.9, 70114.9]]
cr_h5 = "ArepoCosmicRays/snapshot_039.hdf5"


@requires_file(bullet_h5)
def test_arepo_hdf5_selection():
    ds = data_dir_load(bullet_h5)
    assert isinstance(ds, ArepoHDF5Dataset)
    psc = ParticleSelectionComparison(ds)
    psc.run_defaults()


bullet_fields = OrderedDict(
    [
        (("gas", "density"), None),
        (("gas", "temperature"), None),
        (("gas", "temperature"), ("gas", "density")),
        (("gas", "velocity_magnitude"), None),
    ]
)


@requires_ds(bullet_h5)
def test_arepo_bullet():
    ds = data_dir_load(bullet_h5)
    for test in sph_answer(ds, "snapshot_150", 26529600, bullet_fields):
        test_arepo_bullet.__name__ = test.description
        yield test


tng59_fields = OrderedDict(
    [
        (("gas", "density"), None),
        (("gas", "temperature"), None),
        (("gas", "temperature"), ("gas", "density")),
        (("gas", "H_number_density"), None),
        (("gas", "H_p0_number_density"), None),
        (("gas", "H_p1_number_density"), None),
        (("gas", "El_number_density"), None),
        (("gas", "C_number_density"), None),
        (("gas", "velocity_magnitude"), None),
        (("gas", "magnetic_field_strength"), None),
    ]
)


@requires_ds(tng59_h5)
def test_arepo_tng59():
    ds = data_dir_load(tng59_h5, kwargs={"bounding_box": _tng59_bbox})
    for test in sph_answer(ds, "halo_59", 10107142, tng59_fields):
        test_arepo_tng59.__name__ = test.description
        yield test


@requires_ds(tng59_h5)
def test_arepo_tng59_periodicity():
    ds1 = data_dir_load(tng59_h5)
    assert ds1.periodicity == (True, True, True)
    ds2 = data_dir_load(tng59_h5, kwargs={"bounding_box": _tng59_bbox})
    assert ds2.periodicity == (False, False, False)


@requires_ds(tng59_h5)
def test_index_override():
    # This tests that we can supply an index_filename, and that when we do, it
    # doesn't get written if our bounding_box is overwritten.
    tmpfd, tmpname = tempfile.mkstemp(suffix=".ewah")
    os.close(tmpfd)
    ds = data_dir_load(
        tng59_h5, kwargs={"index_filename": tmpname, "bounding_box": _tng59_bbox}
    )
    assert isinstance(ds, ArepoHDF5Dataset)
    ds.index
    assert len(open(tmpname).read()) == 0


@requires_file(tng59_h5)
def test_nh_density():
    ds = data_dir_load(tng59_h5, kwargs={"bounding_box": _tng59_bbox})
    ad = ds.all_data()
    assert_allclose_units(
        ad["gas", "H_number_density"], (ad["gas", "H_nuclei_density"])
    )


@requires_file(tng59_h5)
def test_arepo_tng59_selection():
    ds = data_dir_load(tng59_h5, kwargs={"bounding_box": _tng59_bbox})
    psc = ParticleSelectionComparison(ds)
    psc.run_defaults()


cr_fields = OrderedDict(
    [
        (("gas", "density"), None),
        (("gas", "cosmic_ray_energy_density"), None),
        (("gas", "cosmic_ray_pressure"), None),
    ]
)


@requires_ds(cr_h5)
def test_arepo_cr():
    ds = data_dir_load(cr_h5)
    assert ds.gamma_cr == ds.parameters["GammaCR"]
    for test in sph_answer(ds, "snapshot_039", 28313510, cr_fields):
        test_arepo_cr.__name__ = test.description
        yield test



import matplotlib.pyplot as plt
import numpy as np

import yt

# Enable parallelism in the script (assuming it was called with
# `mpirun -np <n_procs>` )
yt.enable_parallelism()

# By using wildcards such as ? and * with the load command, we can load up a
# Time Series containing all of these datasets simultaneously.
# The "entropy" field that we will use below depends on the electron number
# density, which is not in these datasets by default, so we assume full
# ionization using the "default_species_fields" kwarg.
ts = yt.load(
    "GasSloshingLowRes/sloshing_low_res_hdf5_plt_cnt_0*",
    default_species_fields="ionized",
)

storage = {}

# By using the piter() function, we can iterate on every dataset in
# the TimeSeries object.  By using the storage keyword, we can populate
# a dictionary where the dataset is the key, and sto.result is the value
# for later use when the loop is complete.

# The serial equivalent of piter() here is just "for ds in ts:" .

for store, ds in ts.piter(storage=storage):

    # Create a sphere of radius 100 kpc at the center of the dataset volume
    sphere = ds.sphere("c", (100.0, "kpc"))
    # Calculate the entropy within that sphere
    entr = sphere[("gas", "entropy")].sum()
    # Store the current time and sphere entropy for this dataset in our
    # storage dictionary as a tuple
    store.result = (ds.current_time.in_units("Gyr"), entr)

# Convert the storage dictionary values to a Nx2 array, so the can be easily
# plotted
arr = np.array(list(storage.values()))

# Plot up the results: time versus entropy
plt.semilogy(arr[:, 0], arr[:, 1], "r-")
plt.xlabel("Time (Gyr)")
plt.ylabel("Entropy (ergs/K)")
plt.savefig("time_versus_entropy.png")



"""
Data structures for AdaptaHOP frontend.




"""


import os
import re
from itertools import product
from typing import Optional

import numpy as np

from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer,
)
from yt.data_objects.static_output import Dataset
from yt.frontends.halo_catalog.data_structures import HaloCatalogFile
from yt.funcs import mylog, setdefaultattr
from yt.geometry.particle_geometry_handler import ParticleIndex
from yt.units import Mpc  # type: ignore
from yt.utilities.cython_fortran_utils import FortranFile

from .definitions import ADAPTAHOP_TEMPLATES, ATTR_T, HEADER_ATTRIBUTES
from .fields import AdaptaHOPFieldInfo


class AdaptaHOPParticleIndex(ParticleIndex):
    def _setup_filenames(self):
        template = self.dataset.filename_template
        ndoms = self.dataset.file_count
        cls = self.dataset._file_class
        if ndoms > 1:
            self.data_files = [
                cls(self.dataset, self.io, template % {"num": i}, i, None)
                for i in range(ndoms)
            ]
        else:
            self.data_files = [
                cls(
                    self.dataset,
                    self.io,
                    self.dataset.parameter_filename,
                    0,
                    None,
                )
            ]


class AdaptaHOPDataset(Dataset):
    _index_class = AdaptaHOPParticleIndex
    _file_class = HaloCatalogFile
    _field_info_class = AdaptaHOPFieldInfo

    # AdaptaHOP internally assumes 1Mpc == 3.0824cm
    _code_length_to_Mpc = (1.0 * Mpc).to("cm").value / 3.08e24
    _header_attributes: Optional[ATTR_T] = None
    _halo_attributes: Optional[ATTR_T] = None

    def __init__(
        self,
        filename,
        dataset_type="adaptahop_binary",
        n_ref=16,
        num_zones=2,
        units_override=None,
        unit_system="cgs",
        parent_ds=None,
    ):
        self.n_ref = n_ref
        self.num_zones = num_zones
        if parent_ds is None:
            raise RuntimeError(
                "The AdaptaHOP frontend requires a parent dataset "
                "to be passed as `parent_ds`."
            )
        self.parent_ds = parent_ds

        self._guess_headers_from_file(filename)

        super().__init__(
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
        )

    def _set_code_unit_attributes(self):
        setdefaultattr(self, "length_unit", self.quan(self._code_length_to_Mpc, "Mpc"))
        setdefaultattr(self, "mass_unit", self.quan(1e11, "Msun"))
        setdefaultattr(self, "velocity_unit", self.quan(1.0, "km / s"))
        setdefaultattr(self, "time_unit", self.length_unit / self.velocity_unit)

    def _guess_headers_from_file(self, filename) -> None:
        with FortranFile(filename) as fpu:
            ok = False
            for dp, longint in product((True, False), (True, False)):
                fpu.seek(0)
                try:
                    header_attributes = HEADER_ATTRIBUTES(double=dp, longint=longint)
                    fpu.read_attrs(header_attributes)
                    ok = True
                    break
                except (ValueError, OSError):
                    pass

            if not ok:
                raise OSError("Could not read headers from file %s" % filename)

            istart = fpu.tell()
            fpu.seek(0, 2)
            iend = fpu.tell()

            # Try different templates
            ok = False
            for name, cls in ADAPTAHOP_TEMPLATES.items():
                fpu.seek(istart)
                attributes = cls(longint, dp).HALO_ATTRIBUTES
                mylog.debug("Trying %s(longint=%s, dp=%s)", name, longint, dp)
                try:
                    # Try to read two halos to be sure
                    fpu.read_attrs(attributes)
                    if fpu.tell() < iend:
                        fpu.read_attrs(attributes)
                    ok = True
                    break
                except (ValueError, OSError):
                    continue

        if not ok:
            raise OSError("Could not guess fields from file %s" % filename)

        self._header_attributes = header_attributes
        self._halo_attributes = attributes

    def _parse_parameter_file(self):
        with FortranFile(self.parameter_filename) as fpu:
            params = fpu.read_attrs(self._header_attributes)
        self.dimensionality = 3
        # Domain related things
        self.filename_template = self.parameter_filename
        self.file_count = 1
        nz = self.num_zones
        self.domain_dimensions = np.ones(3, "int32") * nz

        # Set things up
        self.cosmological_simulation = 1
        self.current_redshift = (1.0 / params["aexp"]) - 1.0
        self.omega_matter = params["omega_t"]
        self.current_time = self.quan(params["age"], "Gyr")
        self.omega_lambda = 0.724  # hard coded if not inferred from parent ds
        self.hubble_constant = 0.7  # hard coded if not inferred from parent ds
        self._periodicity = (True, True, True)
        self.particle_types = "halos"
        self.particle_types_raw = "halos"

        # Inherit stuff from parent ds -- if they exist
        for k in ("omega_lambda", "hubble_constant", "omega_matter", "omega_radiation"):
            v = getattr(self.parent_ds, k, None)
            if v is not None:
                setattr(self, k, v)

        self.domain_left_edge = np.array([0.0, 0.0, 0.0])
        self.domain_right_edge = (
            self.parent_ds.domain_right_edge.to("Mpc").value * self._code_length_to_Mpc
        )

        self.parameters.update(params)

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        fname = os.path.split(filename)[1]
        if not fname.startswith("tree_bricks") or not re.match(
            r"^tree_bricks\d{3}$", fname
        ):
            return False
        return True

    def halo(self, halo_id, ptype="DM"):
        """
        Create a data container to get member particles and individual
        values from halos. Halo mass, position, and velocity are set as attributes.
        Halo IDs are accessible through the field, "member_ids".  Other fields that
        are one value per halo are accessible as normal.  The field list for
        halo objects can be seen in `ds.halos_field_list`.

        Parameters
        ----------
        halo_id : int
            The id of the halo or group
        ptype : str, default DM
            The type of particles the halo is made of.
        """
        return AdaptaHOPHaloContainer(
            ptype, halo_id, parent_ds=self.parent_ds, halo_ds=self
        )


class AdaptaHOPHaloContainer(YTSelectionContainer):
    """
    Create a data container to get member particles and individual
    values from halos. Halo mass, position, and velocity are set as attributes.
    Halo IDs are accessible through the field, "member_ids".  Other fields that
    are one value per halo are accessible as normal.  The field list for
    halo objects can be seen in `ds.halos_field_list`.

    Parameters
    ----------
    ptype : string
        The type of halo, either "Group" for the main halo or
        "Subhalo" for subhalos.
    particle_identifier : int or tuple of ints
        The halo or subhalo id.  If requesting a subhalo, the id
        can also be given as a tuple of the main halo id and
        subgroup id, such as (1, 4) for subgroup 4 of halo 1.

    Attributes
    ----------
    particle_identifier : int
        The id of the halo or subhalo.
    group_identifier : int
        For subhalos, the id of the enclosing halo.
    subgroup_identifier : int
        For subhalos, the relative id of the subhalo within
        the enclosing halo.
    particle_number : int
        Number of particles in the halo.
    mass : float
        Halo mass.
    position : array of floats
        Halo position.
    velocity : array of floats
        Halo velocity.

    Note
    ----
    Relevant Fields:

     * particle_number - number of particles
     * subhalo_number - number of subhalos
     * group_identifier - id of parent group for subhalos

    Examples
    --------

    >>> import yt
    >>> ds = yt.load(
    ...     "output_00080_halos/tree_bricks080",
    ...     parent_ds=yt.load("output_00080/info_00080.txt"),
    ... )
    >>> ds.halo(1, ptype="io")
    >>> print(halo.mass)
    119.22804260253906 100000000000.0*Msun
    >>> print(halo.position)
    [26.80901299 24.35978484  5.45388672] code_length
    >>> print(halo.velocity)
    [3306394.95849609 8584366.60766602 9982682.80029297] cm/s
    >>> print(halo["io", "particle_mass"])
    [3.19273578e-06 3.19273578e-06 ... 3.19273578e-06 3.19273578e-06] code_mass

    >>> # particle ids for this halo
    >>> print(halo.member_ids)
    [     48      64     176 ... 999947 1005471 1006779]

    """

    _type_name = "halo"
    _con_args = ("ptype", "particle_identifier", "parent_ds", "halo_ds")
    _spatial = False
    # Do not register it to prevent .halo from being attached to all datasets
    _skip_add = True

    def __init__(self, ptype, particle_identifier, parent_ds, halo_ds):
        if ptype not in parent_ds.particle_types_raw:
            raise RuntimeError(
                'Possible halo types are %s, supplied "%s".'
                % (parent_ds.particle_types_raw, ptype)
            )

        # Setup required fields
        self._dimensionality = 3
        self.ds = parent_ds

        # Halo-specific
        self.halo_ds = halo_ds
        self.ptype = ptype
        self.particle_identifier = particle_identifier

        # Set halo particles data
        self._set_halo_properties()
        self._set_halo_member_data()

        # Call constructor
        super().__init__(parent_ds, {})

    def __repr__(self):
        return "%s_%s_%09d" % (self.ds, self.ptype, self.particle_identifier)

    def __getitem__(self, key):
        return self.region[key]

    @property
    def ihalo(self):
        """The index in the catalog of the halo"""
        ihalo = getattr(self, "_ihalo", None)
        if ihalo:
            return ihalo
        else:
            halo_id = self.particle_identifier
            halo_ids = self.halo_ds.r["halos", "particle_identifier"].astype(int)
            ihalo = np.searchsorted(halo_ids, halo_id)

            assert halo_ids[ihalo] == halo_id

            self._ihalo = ihalo
            return self._ihalo

    def _set_halo_member_data(self):
        ptype = self.ptype
        halo_ds = self.halo_ds
        parent_ds = self.ds
        ihalo = self.ihalo

        # Note: convert to physical units to prevent errors when jumping
        # from halo_ds to parent_ds
        halo_pos = halo_ds.r["halos", "particle_position"][ihalo, :].to("Mpc").value
        halo_vel = halo_ds.r["halos", "particle_velocity"][ihalo, :].to("km/s").value
        halo_radius = halo_ds.r["halos", "r"][ihalo].to("Mpc").value

        members = self.member_ids
        ok = False
        f = 1 / 1.1
        center = parent_ds.arr(halo_pos, "Mpc")
        radius = parent_ds.arr(halo_radius, "Mpc")
        # Find smallest sphere containing all particles
        while not ok:
            f *= 1.1
            sph = parent_ds.sphere(center, f * radius)

            part_ids = sph[ptype, "particle_identity"].astype(int)

            ok = len(np.lib.arraysetops.setdiff1d(members, part_ids)) == 0

        # Set bulk velocity
        sph.set_field_parameter("bulk_velocity", (halo_vel, "km/s"))

        # Build subregion that only contains halo particles
        reg = sph.cut_region(
            ['np.in1d(obj[("io", "particle_identity")].astype(int), members)'],
            locals=dict(members=members, np=np),
        )

        self.sphere = sph
        self.region = reg

    def _set_halo_properties(self):
        ihalo = self.ihalo
        ds = self.halo_ds
        # Add position, mass, velocity member functions
        for attr_name in ("mass", "position", "velocity"):
            setattr(self, attr_name, ds.r["halos", f"particle_{attr_name}"][ihalo])
        # Add members
        self.member_ids = self.halo_ds.index.io.members(ihalo).astype(np.int64)



import yt

# Load the dataset.
ds = yt.load("enzo_tiny_cosmology/RD0009/RD0009")

# Create a volume rendering, which will determine data bounds, use the first
# acceptable field in the field_list, and set up a default transfer function.

# Render and save output images with different levels of sigma clipping.
# Sigma clipping removes the highest intensity pixels in a volume render,
# which affects the overall contrast of the image.
sc = yt.create_scene(ds, field=("gas", "density"))
sc.save("clip_0.png")
sc.save("clip_2.png", sigma_clip=2)
sc.save("clip_4.png", sigma_clip=4)
sc.save("clip_6.png", sigma_clip=6)



import os
import shutil
import tempfile
import unittest

import numpy as np
from nose.tools import assert_raises

from yt.testing import assert_equal, fake_random_ds
from yt.visualization.image_writer import (
    apply_colormap,
    multi_image_composite,
    splat_points,
    write_bitmap,
)


class TestImageWriter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tmpdir = tempfile.mkdtemp()
        cls.curdir = os.getcwd()
        os.chdir(cls.tmpdir)

    @classmethod
    def tearDownClass(cls):
        os.chdir(cls.curdir)
        shutil.rmtree(cls.tmpdir)

    def test_multi_image_composite(self):
        ds = fake_random_ds(64, nprocs=4, particles=16**3)
        center = [0.5, 0.5, 0.5]
        normal = [1, 1, 1]
        cut = ds.cutting(normal, center)
        frb = cut.to_frb((0.75, "unitary"), 64)
        multi_image_composite(
            "multi_channel1.png", frb[("index", "x")], frb[("index", "y")]
        )

        # Test multi_image_composite with user specified scaling values
        mi = ds.quan(0.1, "code_length")
        ma = ds.quan(0.9, "code_length")
        multi_image_composite(
            "multi_channel2.png",
            (frb[("index", "x")], mi, ma),
            [frb[("index", "y")], mi, None],
            green_channel=frb[("index", "z")],
            alpha_channel=frb[("gas", "density")],
        )

        # Test with numpy integer array
        x = np.array(np.random.randint(0, 256, size=(10, 10)), dtype="uint8")
        y = np.array(np.random.randint(0, 256, size=(10, 10)), dtype="uint8")
        multi_image_composite("multi_channel3.png", x, y)

    def test_write_bitmap(self):
        image = np.zeros([16, 16, 4], dtype="uint8")
        xs = np.random.rand(100)
        ys = np.random.rand(100)
        image = splat_points(image, xs, ys)
        png_str = write_bitmap(image.copy(), None)

        image_trans = image.swapaxes(0, 1).copy(order="C")
        png_str_trans = write_bitmap(image_trans, None, transpose=True)
        assert_equal(png_str, png_str_trans)

        with assert_raises(RuntimeError) as ex:
            write_bitmap(np.ones([16, 16]), None)
        desired = "Expecting image array of shape (N,M,3) or (N,M,4), received (16, 16)"
        assert_equal(str(ex.exception)[:50], desired[:50])

    def test_apply_colormap(self):
        x = np.array(np.random.randint(0, 256, size=(10, 10)), dtype="uint8")
        apply_colormap(x, color_bounds=None, cmap_name=None, func=lambda x: x**2)



import itertools

import numpy as np

from yt.testing import assert_almost_equal, fake_amr_ds


def test_amr_kdtree_set_fields():
    ds = fake_amr_ds(fields=["density", "pressure"], units=["g/cm**3", "dyn/cm**2"])
    dd = ds.all_data()

    fields = ds.field_list
    dd.tiles.set_fields(fields, [True, True], False)
    gold = {}
    for i, block in enumerate(dd.tiles.traverse()):
        gold[i] = [data.copy() for data in block.my_data]

    for log_fields in itertools.product([True, False], [True, False]):
        dd.tiles.set_fields(fields, log_fields, False)
        for iblock, block in enumerate(dd.tiles.traverse()):
            for i in range(len(fields)):
                if log_fields[i]:
                    data = block.my_data[i]
                else:
                    data = np.log10(block.my_data[i])
                assert_almost_equal(gold[iblock][i], data)



import numpy as np

from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer
from yt.units.yt_array import YTArray
from yt.utilities.physical_constants import amu_cgs, boltzmann_constant_cgs

b_units = "code_magnetic"
ra_units = "code_length / code_time**2"
rho_units = "code_mass / code_length**3"
vel_units = "code_velocity"
# NOTE: ARTIO uses momentum density.
mom_units = "code_mass / (code_length**2 * code_time)"
en_units = "code_mass*code_velocity**2/code_length**3"
p_units = "code_mass / (code_length * code_time**2)"


class ARTIOFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        ("HVAR_GAS_DENSITY", (rho_units, ["density"], None)),
        ("HVAR_GAS_ENERGY", (en_units, ["total_energy_density"], None)),
        ("HVAR_INTERNAL_ENERGY", (en_units, ["thermal_energy_density"], None)),
        ("HVAR_PRESSURE", (p_units, ["pressure"], None)),
        ("HVAR_MOMENTUM_X", (mom_units, ["momentum_density_x"], None)),
        ("HVAR_MOMENTUM_Y", (mom_units, ["momentum_density_y"], None)),
        ("HVAR_MOMENTUM_Z", (mom_units, ["momentum_density_z"], None)),
        ("HVAR_GAMMA", ("", ["gamma"], None)),
        ("HVAR_METAL_DENSITY_Ia", (rho_units, ["metal_ia_density"], None)),
        ("HVAR_METAL_DENSITY_II", (rho_units, ["metal_ii_density"], None)),
        ("VAR_POTENTIAL", ("", ["potential"], None)),
        ("VAR_POTENTIAL_HYDRO", ("", ["gas_potential"], None)),
        ("RT_HVAR_HI", (rho_units, ["H_density"], None)),
        ("RT_HVAR_HII", (rho_units, ["H_p1_density"], None)),
        ("RT_HVAR_H2", (rho_units, ["H2_density"], None)),
        ("RT_HVAR_HeI", (rho_units, ["He_density"], None)),
        ("RT_HVAR_HeII", (rho_units, ["He_p1_density"], None)),
        ("RT_HVAR_HeIII", (rho_units, ["He_p2_density"], None)),
    )

    known_particle_fields: KnownFieldsT = (
        ("POSITION_X", ("code_length", ["particle_position_x"], None)),
        ("POSITION_Y", ("code_length", ["particle_position_y"], None)),
        ("POSITION_Z", ("code_length", ["particle_position_z"], None)),
        ("VELOCITY_X", (vel_units, ["particle_velocity_x"], None)),
        ("VELOCITY_Y", (vel_units, ["particle_velocity_y"], None)),
        ("VELOCITY_Z", (vel_units, ["particle_velocity_z"], None)),
        ("MASS", ("code_mass", ["particle_mass"], None)),
        ("PID", ("", ["particle_index"], None)),
        ("SPECIES", ("", ["particle_type"], None)),
        ("BIRTH_TIME", ("", [], None)),  # code-units defined as dimensionless to
        # avoid incorrect conversion
        ("INITIAL_MASS", ("code_mass", ["initial_mass"], None)),
        ("METALLICITY_SNIa", ("", ["metallicity_snia"], None)),
        ("METALLICITY_SNII", ("", ["metallicity_snii"], None)),
    )

    def setup_fluid_fields(self):
        unit_system = self.ds.unit_system

        def _get_vel(axis):
            def velocity(field, data):
                return (
                    data[("gas", f"momentum_density_{axis}")] / data[("gas", "density")]
                )

            return velocity

        for ax in "xyz":
            self.add_field(
                ("gas", f"velocity_{ax}"),
                sampling_type="cell",
                function=_get_vel(ax),
                units=unit_system["velocity"],
            )

        def _temperature(field, data):
            tr = data[("gas", "thermal_energy_density")] / data[("gas", "density")]
            # We want this to match *exactly* what ARTIO would compute
            # internally.  We therefore use the exact values that are internal
            # to ARTIO, rather than yt's own internal constants.
            mH = 1.007825 * amu_cgs
            mHe = 4.002602 * amu_cgs
            Yp = 0.24
            XH = 1.0 - Yp
            XHe = 0.25 * Yp
            mb = XH * mH + XHe * mHe
            wmu = 4.0 / (8.0 - 5.0 * Yp)
            # Note that we have gamma = 5.0/3.0 here
            tr *= data[("gas", "gamma")] - 1.0
            tr *= wmu
            tr *= mb / boltzmann_constant_cgs
            return tr

        # TODO: The conversion factor here needs to be addressed, as previously
        # it was set as:
        # unit_T = unit_v**2.0*mb / constants.k
        self.add_field(
            ("gas", "temperature"),
            sampling_type="cell",
            function=_temperature,
            units=unit_system["temperature"],
        )

        # Create a metal_density field as sum of existing metal fields.
        flag1 = ("artio", "HVAR_METAL_DENSITY_Ia") in self.field_list
        flag2 = ("artio", "HVAR_METAL_DENSITY_II") in self.field_list
        if flag1 or flag2:
            if flag1 and flag2:

                def _metal_density(field, data):
                    tr = data[("gas", "metal_ia_density")].copy()
                    np.add(tr, data[("gas", "metal_ii_density")], out=tr)
                    return tr

            elif flag1 and not flag2:

                def _metal_density(field, data):
                    tr = data["metal_ia_density"]
                    return tr

            else:

                def _metal_density(field, data):
                    tr = data["metal_ii_density"]
                    return tr

            self.add_field(
                ("gas", "metal_density"),
                sampling_type="cell",
                function=_metal_density,
                units=unit_system["density"],
                take_log=True,
            )

    def setup_particle_fields(self, ptype):
        if ptype == "STAR":

            def _creation_time(field, data):
                return YTArray(
                    data.ds._handle.tphys_from_tcode_array(data["STAR", "BIRTH_TIME"]),
                    "yr",
                )

            def _age(field, data):
                return data.ds.current_time - data["STAR", "creation_time"]

            self.add_field(
                (ptype, "creation_time"),
                sampling_type="particle",
                function=_creation_time,
                units="yr",
            )
            self.add_field(
                (ptype, "age"), sampling_type="particle", function=_age, units="yr"
            )

            if self.ds.cosmological_simulation:

                def _creation_redshift(field, data):
                    return (
                        1.0
                        / data.ds._handle.auni_from_tcode_array(
                            data["STAR", "BIRTH_TIME"]
                        )
                        - 1.0
                    )

                self.add_field(
                    (ptype, "creation_redshift"),
                    sampling_type="particle",
                    function=_creation_redshift,
                )

        super().setup_particle_fields(ptype)



#!/usr/bin/env python3

import glob
import os
import shutil
import subprocess
import sys
import tempfile
import traceback
from multiprocessing import Pool

import matplotlib

from yt.config import ytcfg

matplotlib.use("Agg")

FPATTERNS = ["*.png", "*.txt", "*.h5", "*.dat", "*.mp4"]
DPATTERNS = ["LC*", "LR", "DD0046"]
BADF = [
    "cloudy_emissivity.h5",
    "apec_emissivity.h5",
    "xray_emissivity.h5",
    "AMRGridData_Slice_x_density.png",
]
CWD = os.getcwd()
ytcfg["yt", "serialize"] = False
BLACKLIST = ["opengl_ipython", "opengl_vr"]


def prep_dirs():
    for directory in glob.glob(f"{ytcfg.get('yt', 'test_data_dir')}/*"):
        os.symlink(directory, os.path.basename(directory))


def run_recipe(payload):
    (recipe,) = payload
    module_name, ext = os.path.splitext(os.path.basename(recipe))
    dest = os.path.join(os.path.dirname(recipe), "_static", module_name)
    if module_name in BLACKLIST:
        return 0
    if not os.path.exists(f"{CWD}/_temp/{module_name}.done"):
        sys.stderr.write(f"Started {module_name}\n")
        tmpdir = tempfile.mkdtemp()
        os.chdir(tmpdir)
        prep_dirs()
        try:
            subprocess.check_call(["python", recipe])
        except Exception:
            trace = "".join(traceback.format_exception(*sys.exc_info()))
            trace += f" in module: {module_name}\n"
            trace += f" recipe: {recipe}\n"
            raise Exception(trace)
        open(f"{CWD}/_temp/{module_name}.done", "wb").close()
        for pattern in FPATTERNS:
            for fname in glob.glob(pattern):
                if fname not in BADF:
                    shutil.move(fname, f"{dest}__{fname}")
        for pattern in DPATTERNS:
            for dname in glob.glob(pattern):
                shutil.move(dname, dest)
        os.chdir(CWD)
        shutil.rmtree(tmpdir, True)
        sys.stderr.write(f"Finished with {module_name}\n")
    return 0


for path in [
    "_temp",
    "source/cookbook/_static",
    "source/visualizing/colormaps/_static",
]:
    fpath = os.path.join(CWD, path)
    if os.path.exists(fpath):
        shutil.rmtree(fpath)
    os.makedirs(fpath)

os.chdir("_temp")
recipes = []
for rpath in ["source/cookbook", "source/visualizing/colormaps"]:
    fpath = os.path.join(CWD, rpath)
    sys.path.append(fpath)
    recipes += glob.glob(f"{fpath}/*.py")
WPOOL = Pool(processes=6)
RES = WPOOL.map_async(run_recipe, ((recipe,) for recipe in recipes))
RES.get()
os.chdir(CWD)




# distutils: libraries = STD_LIBS
"""
Checks for points contained in a volume



"""



import numpy as np

cimport cython
cimport numpy as np
from libc.math cimport sqrt


cdef extern from "math.h":
    double fabs(double x)

@cython.wraparound(False)
@cython.boundscheck(False)
def planar_points_in_volume(
                   np.ndarray[np.float64_t, ndim=2] points,
                   np.ndarray[np.int8_t, ndim=1] pmask,  # pixel mask
                   np.ndarray[np.float64_t, ndim=1] left_edge,
                   np.ndarray[np.float64_t, ndim=1] right_edge,
                   np.ndarray[np.int32_t, ndim=3] mask,
                   float dx):
    cdef np.ndarray[np.int8_t, ndim=1] \
         valid = np.zeros(points.shape[0], dtype='int8')
    cdef int i, dim, count
    cdef int ex
    cdef double dx_inv
    cdef unsigned int idx[3]
    count = 0
    dx_inv = 1.0 / dx
    for i in xrange(points.shape[0]):
        if pmask[i] == 0:
            continue
        ex = 1
        for dim in xrange(3):
            if points[i,dim] < left_edge[dim] or points[i,dim] > right_edge[dim]:
                valid[i] = ex = 0
                break
        if ex == 1:
            for dim in xrange(3):
                idx[dim] = <unsigned int> \
                           ((points[i,dim] - left_edge[dim]) * dx_inv)
            if mask[idx[0], idx[1], idx[2]] == 1:
                valid[i] = 1
                count += 1

    cdef np.ndarray[np.int32_t, ndim=1] result = np.empty(count, dtype='int32')
    count = 0
    for i in xrange(points.shape[0]):
        if valid[i] == 1 and pmask[i] == 1:
            result[count] = i
            count += 1

    return result

cdef inline void set_rotated_pos(
            np.float64_t cp[3], np.float64_t rdds[3][3],
            np.float64_t rorigin[3], int i, int j, int k):
    cdef int oi
    for oi in range(3):
        cp[oi] = rdds[0][oi] * (0.5 + i) \
               + rdds[1][oi] * (0.5 + j) \
               + rdds[2][oi] * (0.5 + k) \
               + rorigin[oi]

#@cython.wraparound(False)
#@cython.boundscheck(False)
def grid_points_in_volume(
                   np.ndarray[np.float64_t, ndim=1] box_lengths,
                   np.ndarray[np.float64_t, ndim=1] box_origin,
                   np.ndarray[np.float64_t, ndim=2] rot_mat,
                   np.ndarray[np.float64_t, ndim=1] grid_left_edge,
                   np.ndarray[np.float64_t, ndim=1] grid_right_edge,
                   np.ndarray[np.float64_t, ndim=1] dds,
                   np.ndarray[np.int32_t, ndim=3] mask,
                   int break_first):
    cdef int n[3]
    cdef int i, j, k
    cdef np.float64_t rds[3][3]
    cdef np.float64_t cur_pos[3]
    cdef np.float64_t rorigin[3]
    for i in range(3):
        rorigin[i] = 0.0
    for i in range(3):
        n[i] = mask.shape[i]
        for j in range(3):
            # Set up our transposed dx, which has a component in every
            # direction
            rds[i][j] = dds[i] * rot_mat[j,i]
            # In our rotated coordinate system, the box origin is 0,0,0
            # so we subtract the box_origin from the grid_origin and rotate
            # that
            rorigin[j] += (grid_left_edge[i] - box_origin[i]) * rot_mat[j,i]

    for i in range(n[0]):
        for j in range(n[1]):
            for k in range(n[2]):
                set_rotated_pos(cur_pos, rds, rorigin, i, j, k)
                if (cur_pos[0] > box_lengths[0]): continue
                if (cur_pos[1] > box_lengths[1]): continue
                if (cur_pos[2] > box_lengths[2]): continue
                if (cur_pos[0] < 0.0): continue
                if (cur_pos[1] < 0.0): continue
                if (cur_pos[2] < 0.0): continue
                if break_first:
                    if mask[i,j,k]: return 1
                else:
                    mask[i,j,k] = 1
    return 0

cdef void normalize_vector(np.float64_t vec[3]):
    cdef int i
    cdef np.float64_t norm = 0.0
    for i in range(3):
        norm += vec[i]*vec[i]
    norm = sqrt(norm)
    for i in range(3):
        vec[i] /= norm

cdef void get_cross_product(np.float64_t v1[3],
                            np.float64_t v2[3],
                            np.float64_t cp[3]):
    cp[0] = v1[1]*v2[2] - v1[2]*v2[1]
    cp[1] = v1[3]*v2[0] - v1[0]*v2[3]
    cp[2] = v1[0]*v2[1] - v1[1]*v2[0]
    #print(cp[0], cp[1], cp[2])

cdef int check_projected_overlap(
        np.float64_t sep_ax[3], np.float64_t sep_vec[3], int gi,
        np.float64_t b_vec[3][3], np.float64_t g_vec[3][3]):
    cdef int g_ax, b_ax
    cdef np.float64_t tba, tga, ba, ga, sep_dot
    ba = ga = sep_dot = 0.0
    for g_ax in range(3):
        # We need the grid vectors, which we'll precompute here
        tba = tga = 0.0
        for b_ax in range(3):
            tba += b_vec[g_ax][b_ax] * sep_vec[b_ax]
            tga += g_vec[g_ax][b_ax] * sep_vec[b_ax]
        ba += fabs(tba)
        ga += fabs(tga)
        sep_dot += sep_vec[g_ax] * sep_ax[g_ax]
    #print(sep_vec[0], sep_vec[1], sep_vec[2],)
    #print(sep_ax[0], sep_ax[1], sep_ax[2])
    return (fabs(sep_dot) > ba+ga)
    # Now we do

@cython.wraparound(False)
@cython.boundscheck(False)
def find_grids_in_inclined_box(
        np.ndarray[np.float64_t, ndim=2] box_vectors,
        np.ndarray[np.float64_t, ndim=1] box_center,
        np.ndarray[np.float64_t, ndim=2] grid_left_edges,
        np.ndarray[np.float64_t, ndim=2] grid_right_edges):

    # http://www.gamasutra.com/view/feature/3383/simple_intersection_tests_for_games.php?page=5
    cdef int n = grid_right_edges.shape[0]
    cdef int g_ax, b_ax, gi
    cdef np.float64_t b_vec[3][3]
    cdef np.float64_t g_vec[3][3]
    cdef np.float64_t a_vec[3][3]
    cdef np.float64_t sep_ax[15][3]
    cdef np.float64_t sep_vec[3]
    cdef np.ndarray[np.int32_t, ndim=1] good = np.zeros(n, dtype='int32')
    cdef np.ndarray[np.float64_t, ndim=2] grid_centers
    # Fill in our axis unit vectors
    for b_ax in range(3):
        for g_ax in range(3):
            a_vec[b_ax][g_ax] = <np.float64_t> (b_ax == g_ax)
    grid_centers = (grid_right_edges + grid_left_edges)/2.0

    # Now we pre-compute our candidate separating axes, because the unit
    # vectors for all the grids are identical
    for b_ax in range(3):
        # We have 6 principal axes we already know, which are the grid (domain)
        # principal axes and the box axes
        sep_ax[b_ax][0] = sep_ax[b_ax][1] = sep_ax[b_ax][2] = 0.0
        sep_ax[b_ax][b_ax] = 1.0 # delta_ijk, for grid axes
        for g_ax in range(3):
            b_vec[b_ax][g_ax] = 0.5*box_vectors[b_ax,g_ax]
            sep_ax[b_ax + 3][g_ax] = b_vec[b_ax][g_ax] # box axes
        normalize_vector(sep_ax[b_ax + 3])
        for g_ax in range(3):
            get_cross_product(b_vec[b_ax], a_vec[g_ax], sep_ax[b_ax*3 + g_ax + 6])
            normalize_vector(sep_ax[b_ax*3 + g_ax + 6])

    for gi in range(n):
        for g_ax in range(3):
            # Calculate the separation vector
            sep_vec[g_ax] = grid_centers[gi, g_ax] - box_center[g_ax]
            # Calculate the grid axis lengths
            g_vec[g_ax][0] = g_vec[g_ax][1] = g_vec[g_ax][2] = 0.0
            g_vec[g_ax][g_ax] = 0.5 * (grid_right_edges[gi, g_ax]
                                     - grid_left_edges[gi, g_ax])
        for b_ax in range(15):
            #print(b_ax,)
            if check_projected_overlap(
                        sep_ax[b_ax], sep_vec, gi,
                        b_vec,  g_vec):
                good[gi] = 1
                break
    return good

def calculate_fill_grids(int fill_level, int refratio, int last_level,
                         np.ndarray[np.int64_t, ndim=1] domain_width,
                         np.ndarray[np.int64_t, ndim=1] cg_start_index,
                         np.ndarray[np.int32_t, ndim=1] cg_dims,
                         np.ndarray[np.int64_t, ndim=1] g_start_index,
                         np.ndarray[np.int32_t, ndim=1] g_dims,
                         np.ndarray[np.uint8_t, ndim=3, cast=True] g_child_mask):
    cdef np.int64_t cgstart[3]
    cdef np.int64_t gstart[3]
    cdef np.int64_t cgend[3]
    cdef np.int64_t gend[3]
    cdef np.int64_t dw[3]
    cdef np.int64_t cxi, cyi, czi, gxi, gyi, gzi, ci, cj, ck
    cdef int i, total = 0
    for i in range(3):
        dw[i] = domain_width[i]
        cgstart[i] = cg_start_index[i]
        gstart[i] = g_start_index[i]
        cgend[i] = cgstart[i] + cg_dims[i]
        gend[i] = gstart[i] + g_dims[i]
    for cxi in range(cgstart[0], cgend[0]+1):
        ci = (cxi % dw[0])
        if ci < 0: ci += dw[0]
        if ci < gstart[0]*refratio or ci >= gend[0]*refratio: continue
        gxi = (<np.int64_t> (ci / refratio)) - gstart[0]
        for cyi in range(cgstart[1], cgend[1]):
            cj = (cyi % dw[1])
            if cj < 0: cj += dw[1]
            if cj < gstart[1]*refratio or cj >= gend[1]*refratio: continue
            gyi = (<np.int64_t> (cj / refratio)) - gstart[1]
            for czi in range(cgstart[2], cgend[2]):
                ck = (czi % dw[2])
                if ck < 0: ck += dw[2]
                if ck < gstart[2]*refratio or cj >= gend[2]*refratio: continue
                gzi = (<np.int64_t> (ck / refratio)) - gstart[2]
                if last_level or g_child_mask[gxi, gyi, gzi] > 0: total += 1
    return total



import os
import shutil
import tempfile
from unittest import TestCase

import yt
from yt.testing import fake_random_ds


def setup():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


class SigmaClipTest(TestCase):
    # This toggles using a temporary directory. Turn off to examine images.
    use_tmpdir = True

    def setUp(self):
        if self.use_tmpdir:
            self.curdir = os.getcwd()
            # Perform I/O in safe place instead of yt main dir
            self.tmpdir = tempfile.mkdtemp()
            os.chdir(self.tmpdir)
        else:
            self.curdir, self.tmpdir = None, None

    def tearDown(self):
        if self.use_tmpdir:
            os.chdir(self.curdir)
            shutil.rmtree(self.tmpdir)

    def test_sigma_clip(self):
        ds = fake_random_ds(32)
        sc = yt.create_scene(ds)
        sc.save("clip_2.png", sigma_clip=2)



import warnings
from typing import Optional


class VisibleDeprecationWarning(UserWarning):
    """Visible deprecation warning, adapted from NumPy

    The nose runner does not show users DeprecationWarning.
    This ensures that a deprecation warning is visible to users
    if that is desired.
    """

    # this class becomes useless after the tests are migrated from nose to pytest

    pass


def issue_deprecation_warning(
    msg: str, *, since: str, removal: Optional[str] = None, stacklevel: int = 3
):
    """
    Parameters
    ----------
    msg : str
        A text message explaining that the code surrounding the call to this function is
        deprecated, and what should be changed on the user side to avoid it.

    since and removal: str version numbers, indicating the anticipated removal date

    Notes
    -----

    removal can be left empty if it is not clear how many minor releases are expected to
    happen before the next major.

    removal and since arguments are keyword-only to forbid accidentally swapping them.

    Examples
    --------
    >>> issue_deprecation_warning(
    ...     "This code is deprecated.", since="4.0.0", removal="4.2.0"
    ... )
    """

    msg += f"\nDeprecated since yt {since}\nThis feature is planned for removal "
    if removal is None:
        msg += "two minor releases later (anticipated)"
    else:
        msg += f"in yt {removal}"
    warnings.warn(msg, VisibleDeprecationWarning, stacklevel=stacklevel)



import os
import weakref
from itertools import chain, product

import numpy as np

from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.index_subobjects.unstructured_mesh import SemiStructuredMesh
from yt.data_objects.static_output import Dataset
from yt.fields.magnetic_field import get_magnetic_normalization
from yt.funcs import get_pbar, mylog
from yt.geometry.grid_geometry_handler import GridIndex
from yt.geometry.unstructured_mesh_handler import UnstructuredIndex
from yt.utilities.chemical_formulas import compute_mu
from yt.utilities.file_handler import HDF5FileHandler

from .fields import AthenaPPFieldInfo

geom_map = {
    "cartesian": "cartesian",
    "cylindrical": "cylindrical",
    "spherical_polar": "spherical",
    "minkowski": "cartesian",
    "tilted": "cartesian",
    "sinusoidal": "cartesian",
    "schwarzschild": "spherical",
    "kerr-schild": "spherical",
}

_cis = np.fromiter(
    chain.from_iterable(product([0, 1], [0, 1], [0, 1])), dtype=np.int64, count=8 * 3
)
_cis.shape = (8, 3)


class AthenaPPLogarithmicMesh(SemiStructuredMesh):
    _index_offset = 0

    def __init__(
        self,
        mesh_id,
        filename,
        connectivity_indices,
        connectivity_coords,
        index,
        blocks,
        dims,
    ):
        super().__init__(
            mesh_id, filename, connectivity_indices, connectivity_coords, index
        )
        self.mesh_blocks = blocks
        self.mesh_dims = dims


class AthenaPPLogarithmicIndex(UnstructuredIndex):
    def __init__(self, ds, dataset_type="athena_pp"):
        self._handle = ds._handle
        super().__init__(ds, dataset_type)
        self.index_filename = self.dataset.filename
        self.directory = os.path.dirname(self.dataset.filename)
        self.dataset_type = dataset_type

    def _initialize_mesh(self):
        mylog.debug("Setting up meshes.")
        num_blocks = self._handle.attrs["NumMeshBlocks"]
        log_loc = self._handle["LogicalLocations"]
        levels = self._handle["Levels"]
        x1f = self._handle["x1f"]
        x2f = self._handle["x2f"]
        x3f = self._handle["x3f"]
        nbx, nby, nbz = tuple(np.max(log_loc, axis=0) + 1)
        nlevel = self._handle.attrs["MaxLevel"] + 1

        nb = np.array([nbx, nby, nbz], dtype="int64")
        self.mesh_factors = np.ones(3, dtype="int64") * ((nb > 1).astype("int") + 1)

        block_grid = -np.ones((nbx, nby, nbz, nlevel), dtype="int64")
        block_grid[log_loc[:, 0], log_loc[:, 1], log_loc[:, 2], levels[:]] = np.arange(
            num_blocks
        )

        block_list = np.arange(num_blocks, dtype="int64")
        bc = []
        for i in range(num_blocks):
            if block_list[i] >= 0:
                ii, jj, kk = log_loc[i]
                neigh = block_grid[ii : ii + 2, jj : jj + 2, kk : kk + 2, levels[i]]
                if np.all(neigh > -1):
                    loc_ids = neigh.transpose().flatten()
                    bc.append(loc_ids)
                    block_list[loc_ids] = -1
                else:
                    bc.append(np.array(i))
                    block_list[i] = -1

        num_meshes = len(bc)

        self.meshes = []
        pbar = get_pbar("Constructing meshes", num_meshes)
        for i in range(num_meshes):
            ob = bc[i][0]
            x = x1f[ob, :]
            y = x2f[ob, :]
            z = x3f[ob, :]
            if nbx > 1:
                x = np.concatenate([x, x1f[bc[i][1], 1:]])
            if nby > 1:
                y = np.concatenate([y, x2f[bc[i][2], 1:]])
            if nbz > 1:
                z = np.concatenate([z, x3f[bc[i][4], 1:]])
            nxm = x.size
            nym = y.size
            nzm = z.size
            coords = np.zeros((nxm, nym, nzm, 3), dtype="float64", order="C")
            coords[:, :, :, 0] = x[:, None, None]
            coords[:, :, :, 1] = y[None, :, None]
            coords[:, :, :, 2] = z[None, None, :]
            coords.shape = (nxm * nym * nzm, 3)
            cycle = np.rollaxis(np.indices((nxm - 1, nym - 1, nzm - 1)), 0, 4)
            cycle.shape = ((nxm - 1) * (nym - 1) * (nzm - 1), 3)
            off = _cis + cycle[:, np.newaxis]
            connectivity = ((off[:, :, 0] * nym) + off[:, :, 1]) * nzm + off[:, :, 2]
            mesh = AthenaPPLogarithmicMesh(
                i,
                self.index_filename,
                connectivity,
                coords,
                self,
                bc[i],
                np.array([nxm - 1, nym - 1, nzm - 1]),
            )
            self.meshes.append(mesh)
            pbar.update(i + 1)
        pbar.finish()
        mylog.debug("Done setting up meshes.")

    def _detect_output_fields(self):
        self.field_list = [("athena_pp", k) for k in self.ds._field_map]


class AthenaPPGrid(AMRGridPatch):
    _id_offset = 0

    def __init__(self, id, index, level):
        AMRGridPatch.__init__(self, id, filename=index.index_filename, index=index)
        self.Parent = None
        self.Children = []
        self.Level = level

    def _setup_dx(self):
        # So first we figure out what the index is.  We don't assume
        # that dx=dy=dz , at least here.  We probably do elsewhere.
        id = self.id - self._id_offset
        LE, RE = self.index.grid_left_edge[id, :], self.index.grid_right_edge[id, :]
        self.dds = self.ds.arr((RE - LE) / self.ActiveDimensions, "code_length")
        if self.ds.dimensionality < 2:
            self.dds[1] = 1.0
        if self.ds.dimensionality < 3:
            self.dds[2] = 1.0
        self.field_data["dx"], self.field_data["dy"], self.field_data["dz"] = self.dds


class AthenaPPHierarchy(GridIndex):

    grid = AthenaPPGrid
    _dataset_type = "athena_pp"
    _data_file = None

    def __init__(self, ds, dataset_type="athena_pp"):
        self.dataset = weakref.proxy(ds)
        self.directory = os.path.dirname(self.dataset.filename)
        self.dataset_type = dataset_type
        # for now, the index file is the dataset!
        self.index_filename = self.dataset.filename
        self._handle = ds._handle
        GridIndex.__init__(self, ds, dataset_type)

    def _detect_output_fields(self):
        self.field_list = [("athena_pp", k) for k in self.dataset._field_map]

    def _count_grids(self):
        self.num_grids = self._handle.attrs["NumMeshBlocks"]

    def _parse_index(self):
        num_grids = self._handle.attrs["NumMeshBlocks"]

        self.grid_left_edge = np.zeros((num_grids, 3), dtype="float64")
        self.grid_right_edge = np.zeros((num_grids, 3), dtype="float64")
        self.grid_dimensions = np.zeros((num_grids, 3), dtype="int32")

        # TODO: In an unlikely case this would use too much memory, implement
        #       chunked read along 1 dim
        x = self._handle["x1f"][:, :]
        y = self._handle["x2f"][:, :]
        z = self._handle["x3f"][:, :]
        mesh_block_size = self._handle.attrs["MeshBlockSize"]

        for i in range(num_grids):
            self.grid_left_edge[i] = np.array(
                [x[i, 0], y[i, 0], z[i, 0]], dtype="float64"
            )
            self.grid_right_edge[i] = np.array(
                [x[i, -1], y[i, -1], z[i, -1]], dtype="float64"
            )
            self.grid_dimensions[i] = mesh_block_size
        levels = self._handle["Levels"][:]

        self.grid_left_edge = self.ds.arr(self.grid_left_edge, "code_length")
        self.grid_right_edge = self.ds.arr(self.grid_right_edge, "code_length")

        self.grids = np.empty(self.num_grids, dtype="object")
        for i in range(num_grids):
            self.grids[i] = self.grid(i, self, levels[i])

        if self.dataset.dimensionality <= 2:
            self.grid_right_edge[:, 2] = self.dataset.domain_right_edge[2]
        if self.dataset.dimensionality == 1:
            self.grid_right_edge[:, 1:] = self.dataset.domain_right_edge[1:]
        self.grid_particle_count = np.zeros([self.num_grids, 1], dtype="int64")

    def _populate_grid_objects(self):
        for g in self.grids:
            g._prepare_grid()
            g._setup_dx()
        self.max_level = self._handle.attrs["MaxLevel"]


class AthenaPPDataset(Dataset):
    _field_info_class = AthenaPPFieldInfo
    _dataset_type = "athena_pp"

    def __init__(
        self,
        filename,
        dataset_type="athena_pp",
        storage_filename=None,
        parameters=None,
        units_override=None,
        unit_system="code",
        default_species_fields=None,
        magnetic_normalization="gaussian",
    ):
        self.fluid_types += ("athena_pp",)
        if parameters is None:
            parameters = {}
        self.specified_parameters = parameters
        if units_override is None:
            units_override = {}
        self._handle = HDF5FileHandler(filename)
        xrat = self._handle.attrs["RootGridX1"][2]
        yrat = self._handle.attrs["RootGridX2"][2]
        zrat = self._handle.attrs["RootGridX3"][2]
        if xrat != 1.0 or yrat != 1.0 or zrat != 1.0:
            self._index_class = AthenaPPLogarithmicIndex
            self.logarithmic = True
        else:
            self._index_class = AthenaPPHierarchy
            self.logarithmic = False
        self._magnetic_factor = get_magnetic_normalization(magnetic_normalization)
        Dataset.__init__(
            self,
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )
        if storage_filename is None:
            storage_filename = self.basename + ".yt"
        self.storage_filename = storage_filename

    def _set_code_unit_attributes(self):
        """
        Generates the conversion to various physical _units based on the
        parameter file
        """
        if "length_unit" not in self.units_override:
            self.no_cgs_equiv_length = True
        for unit, cgs in [
            ("length", "cm"),
            ("time", "s"),
            ("mass", "g"),
            ("temperature", "K"),
        ]:
            # We set these to cgs for now, but they may have been overridden
            if getattr(self, unit + "_unit", None) is not None:
                continue
            mylog.warning("Assuming 1.0 = 1.0 %s", cgs)
            setattr(self, f"{unit}_unit", self.quan(1.0, cgs))

        self.magnetic_unit = np.sqrt(
            self._magnetic_factor
            * self.mass_unit
            / (self.time_unit**2 * self.length_unit)
        )
        self.magnetic_unit.convert_to_units("gauss")
        self.velocity_unit = self.length_unit / self.time_unit

    def _parse_parameter_file(self):

        xmin, xmax = self._handle.attrs["RootGridX1"][:2]
        ymin, ymax = self._handle.attrs["RootGridX2"][:2]
        zmin, zmax = self._handle.attrs["RootGridX3"][:2]

        self.domain_left_edge = np.array([xmin, ymin, zmin], dtype="float64")
        self.domain_right_edge = np.array([xmax, ymax, zmax], dtype="float64")

        self.geometry = geom_map[self._handle.attrs["Coordinates"].decode("utf-8")]
        self.domain_width = self.domain_right_edge - self.domain_left_edge
        self.domain_dimensions = self._handle.attrs["RootGridSize"]

        self._field_map = {}
        k = 0
        for dname, num_var in zip(
            self._handle.attrs["DatasetNames"], self._handle.attrs["NumVariables"]
        ):
            for j in range(num_var):
                fname = self._handle.attrs["VariableNames"][k].decode("ascii", "ignore")
                self._field_map[fname] = (dname.decode("ascii", "ignore"), j)
                k += 1

        self.refine_by = 2
        dimensionality = 3
        if self.domain_dimensions[2] == 1:
            dimensionality = 2
        if self.domain_dimensions[1] == 1:
            dimensionality = 1
        self.dimensionality = dimensionality
        self.current_time = self._handle.attrs["Time"]
        self.cosmological_simulation = False
        self.num_ghost_zones = 0
        self.field_ordering = "fortran"
        self.boundary_conditions = [1] * 6
        self._periodicity = tuple(
            self.specified_parameters.get("periodicity", (True, True, True))
        )
        if "gamma" in self.specified_parameters:
            self.gamma = float(self.specified_parameters["gamma"])
        else:
            self.gamma = 5.0 / 3.0

        self.current_redshift = 0.0
        self.omega_lambda = 0.0
        self.omega_matter = 0.0
        self.hubble_constant = 0.0
        self.cosmological_simulation = 0
        self.parameters["Time"] = self.current_time  # Hardcode time conversion for now.
        self.parameters[
            "HydroMethod"
        ] = 0  # Hardcode for now until field staggering is supported.
        if "gamma" in self.specified_parameters:
            self.parameters["Gamma"] = self.specified_parameters["gamma"]
        else:
            self.parameters["Gamma"] = 5.0 / 3.0
        self.mu = self.specified_parameters.get(
            "mu", compute_mu(self.default_species_fields)
        )

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        try:
            if filename.endswith(".athdf"):
                return True
        except Exception:
            pass
        return False

    @property
    def _skip_cache(self):
        return True

    def __str__(self):
        return self.basename.rsplit(".", 1)[0]



import yt

# Load the dataset.
ds = yt.load("GalaxyClusterMerger/fiducial_1to3_b0.273d_hdf5_plt_cnt_0175")

# Create projections of temperature (with different methods)


for method in ["integrate", "min", "max"]:
    proj = yt.ProjectionPlot(ds, "x", ("gas", "temperature"), method=method)
    proj.save(f"projection_method_{method}.png")



import numpy as np

import yt

# Load the dataset.  We'll work with a some Gadget data to illustrate all
# the different ways in which the effective resolution can vary.  Specifically,
# we'll use the GadgetDiskGalaxy dataset available at
#  http://yt-project.org/data/GadgetDiskGalaxy.tar.gz

# load the data with a refinement criteria of 2 particle per cell
# n.b. -- in yt-4.0, n_ref no longer exists as the data is no longer
# deposited only a grid.  At present (03/15/2019), there is no way to
# handle non-gas data in Gadget snapshots, though that is work in progress
if int(yt.__version__[0]) < 4:
    # increasing n_ref will result in a "lower resolution" (but faster) image,
    # while decreasing it will go the opposite way
    ds = yt.load("GadgetDiskGalaxy/snapshot_200.hdf5", n_ref=16)
else:
    ds = yt.load("GadgetDiskGalaxy/snapshot_200.hdf5")

# Create projections of the density (max value in each resolution element in the image):
prj = yt.ProjectionPlot(
    ds, "x", ("gas", "density"), method="max", center="max", width=(100, "kpc")
)

# nicen up the plot by using a better interpolation:
plot = prj.plots[list(prj.plots)[0]]
ax = plot.axes
img = ax.images[0]
img.set_interpolation("bicubic")

# nicen up the plot by setting the background color to the minimum of the colorbar
prj.set_background_color(("gas", "density"))

# vary the buff_size -- the number of resolution elements in the actual visualization
# set it to 2000x2000
buff_size = 2000
prj.set_buff_size(buff_size)

# set the figure size in inches
figure_size = 10
prj.set_figure_size(figure_size)

# if the image does not fill the plot (as is default, since the axes and
# colorbar contribute as well), then figuring out the proper dpi for a given
# buff_size and figure_size is non-trivial -- it requires finding the bbox
# for the actual image:
bounding_box = ax.get_position()
# we're going to scale to the larger of the two sides
image_size = figure_size * max([bounding_box.width, bounding_box.height])
# now save with a dpi that's scaled to the buff_size:
dpi = np.rint(np.ceil(buff_size / image_size))
prj.save("with_axes_colorbar.png", mpl_kwargs=dict(dpi=dpi))

# in the case where the image fills the entire plot (i.e. if the axes and colorbar
# are turned off), it's trivial to figure out the correct dpi from the buff_size and
# figure_size (or vice versa):

# hide the colorbar:
prj.hide_colorbar()

# hide the axes, while still keeping the background color correct:
prj.hide_axes(draw_frame=True)

# save with a dpi that makes sense:
dpi = np.rint(np.ceil(buff_size / figure_size))
prj.save("no_axes_colorbar.png", mpl_kwargs=dict(dpi=dpi))



import numpy as np

import yt

# Load the dataset.
ds = yt.load("GasSloshing/sloshing_nomag2_hdf5_plt_cnt_0150")

# Create a slice object
slc = yt.SlicePlot(ds, "x", ("gas", "density"), width=(800.0, "kpc"))

# Get a reference to the matplotlib axes object for the plot
ax = slc.plots[("gas", "density")].axes

# Let's adjust the x axis tick labels
for label in ax.xaxis.get_ticklabels():
    label.set_color("red")
    label.set_fontsize(16)

# Get a reference to the matplotlib figure object for the plot
fig = slc.plots[("gas", "density")].figure

# And create a mini-panel of a gaussian histogram inside the plot
rect = (0.2, 0.2, 0.2, 0.2)
new_ax = fig.add_axes(rect)

n, bins, patches = new_ax.hist(
    np.random.randn(1000) + 20, 50, facecolor="black", edgecolor="black"
)

# Make sure its visible
new_ax.tick_params(colors="white")

# And label it
la = new_ax.set_xlabel("Dinosaurs per furlong")
la.set_color("white")

slc.save()



from typing import List

import numpy as np

from yt.data_objects.profiles import create_profile
from yt.data_objects.static_output import Dataset
from yt.funcs import fix_axis, iter_fields
from yt.units.yt_array import YTArray
from yt.visualization.fixed_resolution import ParticleImageBuffer
from yt.visualization.profile_plotter import PhasePlot

from .plot_window import PWViewerMPL, get_axes_unit, get_window_parameters


class ParticleAxisAlignedDummyDataSource:
    _type_name = "Particle"
    _dimensionality = 2
    _con_args = ("center", "axis", "width", "fields", "weight_field")
    _tds_attrs = ()
    _key_fields: List[str] = []

    def __init__(
        self,
        center,
        ds,
        axis,
        width,
        fields,
        weight_field=None,
        field_parameters=None,
        data_source=None,
        deposition="ngp",
        density=False,
    ):
        self.center = center
        self.ds = ds
        self.axis = axis
        self.width = width
        self.weight_field = weight_field
        self.deposition = deposition
        self.density = density

        if field_parameters is None:
            self.field_parameters = {}
        else:
            self.field_parameters = field_parameters

        LE = center - 0.5 * YTArray(width)
        RE = center + 0.5 * YTArray(width)
        for ax in range(3):
            if not ds.periodicity[ax]:
                LE[ax] = max(LE[ax], ds.domain_left_edge[ax])
                RE[ax] = min(RE[ax], ds.domain_right_edge[ax])

        self.dd = ds.region(
            center,
            LE,
            RE,
            fields,
            field_parameters=field_parameters,
            data_source=data_source,
        )

        fields = self.dd._determine_fields(fields)
        self.fields = fields

    def _determine_fields(self, *args):
        return self.dd._determine_fields(*args)

    def get_field_parameter(self, name, default=None):
        """
        This is typically only used by derived field functions, but
        it returns parameters used to generate fields.
        """
        if name in self.field_parameters:
            return self.field_parameters[name]
        else:
            return default


class ParticleProjectionPlot(PWViewerMPL):
    r"""Creates a particle plot from a dataset

    Given a ds object, an axis to slice along, and a field name
    string, this will return a PWViewerMPL object containing
    the plot.

    The plot can be updated using one of the many helper functions
    defined in PlotWindow.

    Parameters
    ----------
    ds : `Dataset`
         This is the dataset object corresponding to the
         simulation output to be plotted.
    axis : int or one of 'x', 'y', 'z'
         An int corresponding to the axis to slice along (0=x, 1=y, 2=z)
         or the axis name itself
    fields : string, list or None
         If a string or list, the name of the particle field(s) to be used
         on the colorbar. The color shown will correspond to the sum of the
         given field along the line of sight. If None, the particle positions
         will be indicated using a fixed color, instead. Default is None.
    color : 'b', 'g', 'r', 'c', 'm', 'y', 'k', or 'w'
         One the matplotlib-recognized color strings.
         The color that will indicate the particle locations
         on the mesh. This argument is ignored if z_fields is
         not None. Default is 'b'.
    center : A sequence of floats, a string, or a tuple.
         The coordinate of the center of the image. If set to 'c', 'center' or
         left blank, the plot is centered on the middle of the domain. If set to
         'max' or 'm', the center will be located at the maximum of the
         ('gas', 'density') field. Centering on the max or min of a specific
         field is supported by providing a tuple such as ("min","temperature") or
         ("max","dark_matter_density"). Units can be specified by passing in *center*
         as a tuple containing a coordinate and string unit name or by passing
         in a YTArray. If a list or unitless array is supplied, code units are
         assumed.
    width : tuple or a float.
         Width can have four different formats to support windows with variable
         x and y widths.  They are:

         ==================================     =======================
         format                                 example
         ==================================     =======================
         (float, string)                        (10,'kpc')
         ((float, string), (float, string))     ((10,'kpc'),(15,'kpc'))
         float                                  0.2
         (float, float)                         (0.2, 0.3)
         ==================================     =======================

         For example, (10, 'kpc') requests a plot window that is 10 kiloparsecs
         wide in the x and y directions, ((10,'kpc'),(15,'kpc')) requests a
         window that is 10 kiloparsecs wide along the x axis and 15
         kiloparsecs wide along the y axis.  In the other two examples, code
         units are assumed, for example (0.2, 0.3) requests a plot that has an
         x width of 0.2 and a y width of 0.3 in code units.  If units are
         provided the resulting plot axis labels will use the supplied units.
    depth : A tuple or a float
         A tuple containing the depth to project through and the string
         key of the unit: (width, 'unit').  If set to a float, code units
         are assumed. Defaults to the entire domain.
    weight_field : string
         The name of the weighting field.  Set to None for no weight.
         If given, the plot will show a weighted average along the line of
         sight of the fields given in the ``fields`` argument.
    axes_unit : A string
         The name of the unit for the tick labels on the x and y axes.
         Defaults to None, which automatically picks an appropriate unit.
         If axes_unit is '1', 'u', or 'unitary', it will not display the
         units, and only show the axes name.
    origin : string or length 1, 2, or 3 sequence of strings
         The location of the origin of the plot coordinate system.  This is
         represented by '-' separated string or a tuple of strings.  In the
         first index the y-location is given by 'lower', 'upper', or 'center'.
         The second index is the x-location, given as 'left', 'right', or
         'center'.  Finally, the whether the origin is applied in 'domain'
         space, plot 'window' space or 'native' simulation coordinate system
         is given. For example, both 'upper-right-domain' and ['upper',
         'right', 'domain'] both place the origin in the upper right hand
         corner of domain space. If x or y are not given, a value is inferred.
         For instance, 'left-domain' corresponds to the lower-left hand corner
         of the simulation domain, 'center-domain' corresponds to the center
         of the simulation domain, or 'center-window' for the center of the
         plot window. Further examples:

         ==================================     ============================
         format                                 example
         ==================================     ============================
         '{space}'                              'domain'
         '{xloc}-{space}'                       'left-window'
         '{yloc}-{space}'                       'upper-domain'
         '{yloc}-{xloc}-{space}'                'lower-right-window'
         ('{space}',)                           ('window',)
         ('{xloc}', '{space}')                  ('right', 'domain')
         ('{yloc}', '{space}')                  ('lower', 'window')
         ('{yloc}', '{xloc}', '{space}')        ('lower', 'right', 'window')
         ==================================     ============================
    fontsize : integer
         The size of the fonts for the axis, colorbar, and tick labels.
    field_parameters : dictionary
         A dictionary of field parameters than can be accessed by derived
         fields.
    window_size : float
        The size of the window on the longest axis (in units of inches),
        including the margins but not the colorbar.
    aspect : float
         The aspect ratio of the plot.  Set to None for 1.
    data_source : YTSelectionContainer Object
         Object to be used for data selection.  Defaults to a region covering
         the entire simulation.
    deposition : string
        Controls the order of the interpolation of the particles onto the
        mesh. "ngp" is 0th-order "nearest-grid-point" method (the default),
        "cic" is 1st-order "cloud-in-cell".
    density : boolean
        If True, the quantity to be projected will be divided by the area of
        the cells, to make a projected density of the quantity. The plot
        name and units will also reflect this. Default: False

    Examples
    --------

    This will save an image the the file
    'galaxy0030_Particle_z_particle_mass.png'

    >>> from yt import load
    >>> ds = load("IsolatedGalaxy/galaxy0030/galaxy0030")
    >>> p = yt.ParticleProjectionPlot(ds, 2, "particle_mass")
    >>> p.save()

    """
    _plot_type = "Particle"
    _frb_generator = ParticleImageBuffer

    def __init__(
        self,
        ds,
        axis,
        fields=None,
        color="b",
        center="c",
        width=None,
        depth=(1, "1"),
        weight_field=None,
        axes_unit=None,
        origin="center-window",
        fontsize=18,
        field_parameters=None,
        window_size=8.0,
        aspect=None,
        data_source=None,
        deposition="ngp",
        density=False,
    ):
        # this will handle time series data and controllers
        ts = self._initialize_dataset(ds)
        self.ts = ts
        ds = self.ds = ts[0]
        axis = fix_axis(axis, ds)
        (bounds, center, display_center) = get_window_parameters(
            axis, center, width, ds
        )
        if field_parameters is None:
            field_parameters = {}

        if axes_unit is None:
            axes_unit = get_axes_unit(width, ds)

        # if no fields are passed in, we simply mark the x and
        # y fields using a given color. Use the 'particle_ones'
        # field to do this. We also turn off the colorbar in
        # this case.
        self._use_cbar = True
        splat_color = None
        if fields is None:
            fields = [("all", "particle_ones")]
            weight_field = ("all", "particle_ones")
            self._use_cbar = False
            splat_color = color

        depth = ds.coordinates.sanitize_depth(depth)

        x_coord = ds.coordinates.x_axis[axis]
        y_coord = ds.coordinates.y_axis[axis]

        width = np.zeros_like(center)
        width[x_coord] = bounds[1] - bounds[0]
        width[y_coord] = bounds[3] - bounds[2]
        width[axis] = depth[0].in_units(width[x_coord].units)

        ParticleSource = ParticleAxisAlignedDummyDataSource(
            center,
            ds,
            axis,
            width,
            fields,
            weight_field,
            field_parameters=field_parameters,
            data_source=data_source,
            deposition=deposition,
            density=density,
        )

        self.projected = weight_field is None

        PWViewerMPL.__init__(
            self,
            ParticleSource,
            bounds,
            origin=origin,
            fontsize=fontsize,
            fields=fields,
            window_size=window_size,
            aspect=aspect,
            splat_color=splat_color,
            geometry=ds.geometry,
        )

        self.set_axes_unit(axes_unit)

        if not self._use_cbar:
            self.hide_colorbar()


class ParticlePhasePlot(PhasePlot):
    r"""
    Create a 2d particle phase plot from a data source or from
    a `yt.data_objects.profiles.ParticleProfile` object.

    Given a data object (all_data, region, sphere, etc.), an x field,
    y field, and z field (or fields), this will create a particle plot
    by depositing the particles onto a two-dimensional mesh, using either
    nearest grid point or cloud-in-cell deposition.

    Parameters
    ----------
    data_source : YTSelectionContainer or Dataset
        The data object to be profiled, such as all_data, region, or
        sphere. If data_source is a Dataset, data_source.all_data() will be used.
    x_field : str
        The x field for the mesh.
    y_field : str
        The y field for the mesh.
    z_fields : None, str, or list
        If None, particles will be splatted onto the mesh,
        but no colormap will be used.
        If str or list, the name of the field or fields to
        be displayed on the colorbar. The displayed values will
        correspond to the sum of the field or fields along the
        line of sight.
        Default: None.
    color : 'b', 'g', 'r', 'c', 'm', 'y', 'k', or 'w'
        One the matplotlib-recognized color strings.
        The color that will indicate the particle locations
        on the mesh. This argument is ignored if z_fields is
        not None.
        Default : 'b'
    x_bins : int
        The number of bins in x field for the mesh.
        Default: 800.
    y_bins : int
        The number of bins in y field for the mesh.
        Default: 800.
    weight_field : str
        The field to weight by. If given, the plot will show a weighted
        average along the line of sight of the fields given in the
        ``z_fields`` argument. Default: None.
    deposition : str
        Either 'ngp' or 'cic'. Controls what type of
        interpolation will be used to deposit the
        particle z_fields onto the mesh.
        Default: 'ngp'
    fontsize: int
        Font size for all text in the plot.
        Default: 18.
    figure_size : int
        Size in inches of the image.
        Default: 8 (8x8)
    shading : str
        This argument is directly passed down to matplotlib.axes.Axes.pcolormesh
        see
        https://matplotlib.org/3.3.1/gallery/images_contours_and_fields/pcolormesh_grids.html#sphx-glr-gallery-images-contours-and-fields-pcolormesh-grids-py  # noqa
        Default: 'nearest'

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
    >>> ad = ds.all_data()
    >>> plot = ParticlePhasePlot(
    ...     ad,
    ...     "particle_position_x",
    ...     "particle_position_y",
    ...     ["particle_mass"],
    ...     x_bins=800,
    ...     y_bins=800,
    ... )
    >>> plot.save()

    >>> # Change plot properties.
    >>> plot.set_log("particle_mass", True)
    >>> plot.set_unit("particle_position_x", "Mpc")
    >>> plot.set_unit("particle_velocity_z", "km/s")
    >>> plot.set_unit("particle_mass", "Msun")

    """
    _plot_type = "ParticlePhase"

    def __init__(
        self,
        data_source,
        x_field,
        y_field,
        z_fields=None,
        color="b",
        x_bins=800,
        y_bins=800,
        weight_field=None,
        deposition="ngp",
        fontsize=18,
        figure_size=8.0,
        shading="nearest",
    ):

        if isinstance(data_source, Dataset):
            data_source = data_source.all_data()
        # if no z_fields are passed in, use a constant color
        if z_fields is None:
            self.use_cbar = False
            self.splat_color = color
            z_fields = [("all", "particle_ones")]

        profile = create_profile(
            data_source,
            [x_field, y_field],
            list(iter_fields(z_fields)),
            n_bins=[x_bins, y_bins],
            weight_field=weight_field,
            deposition=deposition,
        )

        type(self)._initialize_instance(
            self, data_source, profile, fontsize, figure_size, shading
        )


def ParticlePlot(ds, x_field, y_field, z_fields=None, color="b", *args, **kwargs):
    r"""
    A factory function for
    :class:`yt.visualization.particle_plots.ParticleProjectionPlot`
    and :class:`yt.visualization.profile_plotter.ParticlePhasePlot` objects.
    This essentially allows for a single entry point to both types of particle
    plots, the distinction being determined by the fields passed in.

    If the x_field and y_field combination corresponds to a valid, right-handed
    spatial plot, an ``ParticleProjectionPlot`` will be returned. This plot
    object can be updated using one of the many helper functions defined in
    ``PlotWindow``.

    If the x_field and y_field combo do not correspond to a valid
    ``ParticleProjectionPlot``, then a ``ParticlePhasePlot``. This object can be
    modified by its own set of  helper functions defined in PhasePlot. We note
    below which arguments are only accepted by ``ParticleProjectionPlot`` and
    which arguments are only accepted by ``ParticlePhasePlot``.

    Parameters
    ----------

    ds : :class:`yt.data_objects.static_output.Dataset`
        This is the dataset object corresponding to the
        simulation output to be plotted.
    x_field : string
        This is the particle field that will be plotted on the x-axis.
    y_field : string
        This is the particle field that will be plotted on the y-axis.
    z_fields : string, list, or None.
        If None, particles will be splatted onto the plot, but no colormap
        will be used. The particle color will instead be determined by
        the 'color' argument. If str or list, the name of the field or fields
        to be displayed on the colorbar.
        Default: None.
    color : 'b', 'g', 'r', 'c', 'm', 'y', 'k', or 'w'
         One the matplotlib-recognized color strings.
         The color that will indicate the particle locations
         on the plot. This argument is ignored if z_fields is
         not None. Default is 'b'.
    weight_field : string
         The name of the weighting field.  Set to None for no weight.
    fontsize : integer
         The size of the fonts for the axis, colorbar, and tick labels.
    data_source : YTSelectionContainer Object
         Object to be used for data selection.  Defaults to a region covering
         the entire simulation.
    center : A sequence of floats, a string, or a tuple.
         The coordinate of the center of the image. If set to 'c', 'center' or
         left blank, the plot is centered on the middle of the domain. If set to
         'max' or 'm', the center will be located at the maximum of the
         ('gas', 'density') field. Centering on the max or min of a specific
         field is supported by providing a tuple such as ("min","temperature") or
         ("max","dark_matter_density"). Units can be specified by passing in *center*
         as a tuple containing a coordinate and string unit name or by passing
         in a YTArray. If a list or unitless array is supplied, code units are
         assumed. This argument is only accepted by ``ParticleProjectionPlot``.
    width : tuple or a float.
         Width can have four different formats to support windows with variable
         x and y widths.  They are:

         ==================================     =======================
         format                                 example
         ==================================     =======================
         (float, string)                        (10,'kpc')
         ((float, string), (float, string))     ((10,'kpc'),(15,'kpc'))
         float                                  0.2
         (float, float)                         (0.2, 0.3)
         ==================================     =======================

         For example, (10, 'kpc') requests a plot window that is 10 kiloparsecs
         wide in the x and y directions, ((10,'kpc'),(15,'kpc')) requests a
         window that is 10 kiloparsecs wide along the x axis and 15
         kiloparsecs wide along the y axis.  In the other two examples, code
         units are assumed, for example (0.2, 0.3) requests a plot that has an
         x width of 0.2 and a y width of 0.3 in code units.  If units are
         provided the resulting plot axis labels will use the supplied units.
         This argument is only accepted by ``ParticleProjectionPlot``.
    depth : A tuple or a float
         A tuple containing the depth to project through and the string
         key of the unit: (width, 'unit').  If set to a float, code units
         are assumed. Defaults to the entire domain. This argument is only
         accepted by ``ParticleProjectionPlot``.
    axes_unit : A string
         The name of the unit for the tick labels on the x and y axes.
         Defaults to None, which automatically picks an appropriate unit.
         If axes_unit is '1', 'u', or 'unitary', it will not display the
         units, and only show the axes name.
    origin : string or length 1, 2, or 3 sequence of strings
         The location of the origin of the plot coordinate system.  This is
         represented by '-' separated string or a tuple of strings.  In the
         first index the y-location is given by 'lower', 'upper', or 'center'.
         The second index is the x-location, given as 'left', 'right', or
         'center'.  Finally, the whether the origin is applied in 'domain'
         space, plot 'window' space or 'native' simulation coordinate system
         is given. For example, both 'upper-right-domain' and ['upper',
         'right', 'domain'] both place the origin in the upper right hand
         corner of domain space. If x or y are not given, a value is inferred.
         For instance, 'left-domain' corresponds to the lower-left hand corner
         of the simulation domain, 'center-domain' corresponds to the center
         of the simulation domain, or 'center-window' for the center of the
         plot window. Further examples:

         ==================================     ============================
         format                                 example
         ==================================     ============================
         '{space}'                              'domain'
         '{xloc}-{space}'                       'left-window'
         '{yloc}-{space}'                       'upper-domain'
         '{yloc}-{xloc}-{space}'                'lower-right-window'
         ('{space}',)                           ('window',)
         ('{xloc}', '{space}')                  ('right', 'domain')
         ('{yloc}', '{space}')                  ('lower', 'window')
         ('{yloc}', '{xloc}', '{space}')        ('lower', 'right', 'window')
         ==================================     ============================

         This argument is only accepted by ``ParticleProjectionPlot``.
    window_size : float
         The size of the window on the longest axis (in units of inches),
         including the margins but not the colorbar. This argument is only
         accepted by ``ParticleProjectionPlot``.
    aspect : float
         The aspect ratio of the plot.  Set to None for 1. This argument is
         only accepted by ``ParticleProjectionPlot``.
    x_bins : int
        The number of bins in x field for the mesh. Defaults to 800. This
        argument is only accepted by ``ParticlePhasePlot``.
    y_bins : int
        The number of bins in y field for the mesh. Defaults to 800. This
        argument is only accepted by ``ParticlePhasePlot``.
    deposition : str
        Either 'ngp' or 'cic'. Controls what type of interpolation will be
        used to deposit the particle z_fields onto the mesh. Defaults to 'ngp'.
    figure_size : int
        Size in inches of the image. Defaults to 8 (product an 8x8 inch figure).
        This argument is only accepted by ``ParticlePhasePlot``.

    Examples
    --------

    >>> from yt import load
    >>> ds = load("IsolatedGalaxy/galaxy0030/galaxy0030")
    >>> p = yt.ParticlePlot(
    ...     ds,
    ...     "particle_position_x",
    ...     "particle_position_y",
    ...     "particle_mass",
    ...     width=(0.5, 0.5),
    ... )
    >>> p.set_unit("particle_mass", "Msun")
    >>> p = yt.ParticlePlot(ds, "particle_position_x", "particle_velocity_z", color="g")

    """
    dd = kwargs.get("data_source", None)
    if dd is None:
        dd = ds.all_data()
    x_field = dd._determine_fields(x_field)[0]
    y_field = dd._determine_fields(y_field)[0]

    direction = 3
    # try potential axes for a ParticleProjectionPlot:
    for axis in [0, 1, 2]:
        xax = ds.coordinates.x_axis[axis]
        yax = ds.coordinates.y_axis[axis]
        ax_field_template = "particle_position_%s"
        xf = ax_field_template % ds.coordinates.axis_name[xax]
        yf = ax_field_template % ds.coordinates.axis_name[yax]
        if (x_field[1], y_field[1]) in [(xf, yf), (yf, xf)]:
            direction = axis
            break

    if direction < 3:
        # Make a ParticleProjectionPlot
        return ParticleProjectionPlot(ds, direction, z_fields, color, *args, **kwargs)

    # Does not correspond to any valid PlotWindow-style plot,
    # use ParticlePhasePlot instead
    else:
        return ParticlePhasePlot(dd, x_field, y_field, z_fields, color, *args, **kwargs)



from nose.plugins.attrib import attr

from yt.testing import ANSWER_TEST_TAG, fake_random_ds
from yt.utilities.answer_testing.framework import GenericImageTest
from yt.visualization.api import LinePlot


@attr(ANSWER_TEST_TAG)
def test_lineplot_set_axis_properties():
    ds = fake_random_ds(16)

    def create_image(filename_prefix):
        p = LinePlot(
            ds,
            ("gas", "density"),
            start_point=[0, 0, 0],
            end_point=[1, 1, 1],
            npoints=32,
        )
        p.set_x_unit("cm")
        p.save(f"{filename_prefix}_xunit")

        p.set_unit(("gas", "density"), "kg/cm**3")
        p.save(f"{filename_prefix}_xunit_zunit")

        p.set_log(("gas", "density"), False)
        p.save(f"{filename_prefix}_xunit_zunit_lin")

    test = GenericImageTest(ds, create_image, 12)
    test.prefix = "test_lineplot_set_axis_properties"
    test.answer_name = "lineplot_set_axis_properties"
    yield test



from yt.testing import assert_equal
from yt.utilities.periodic_table import _elements, periodic_table


def test_element_accuracy():
    for num, w, name, sym in _elements:
        e0 = periodic_table[num]
        e1 = periodic_table[name]
        e2 = periodic_table[sym]
        # If num == -1, then we are in one of the things like Deuterium or El
        # that are not elements by themselves.
        if num == -1:
            e0 = e1
        assert_equal(id(e0), id(e1))
        assert_equal(id(e0), id(e2))
        assert_equal(e0.num, num)
        assert_equal(e0.weight, w)
        assert_equal(e0.name, name)
        assert_equal(e0.symbol, sym)



from .domain_context import DomainContext

# Here's how this all works:
#
#   1. We have a mapping here that defines fields we might expect to find in an
#      astrophysical simulation to units (not necessarily definitive) that we
#      may want to use them in.
#   2. Simulations and frontends will register aliases from fields (which can
#      utilize units) to the fields enumerated here.
#   3. This plugin can call additional plugins on the registry.


class AstroSimulation(DomainContext):

    # This is an immutable of immutables.  Note that we do not specify the
    # fluid type here, although in most cases we expect it to be "gas".
    _known_fluid_fields = (
        ("density", "g/cm**3"),
        ("number_density", "1/cm**3"),
        ("pressure", "dyne / cm**2"),
        ("specific_thermal_energy", "erg / g"),
        ("temperature", "K"),
        ("velocity_x", "cm / s"),
        ("velocity_y", "cm / s"),
        ("velocity_z", "cm / s"),
        ("magnetic_field_x", "gauss"),
        ("magnetic_field_y", "gauss"),
        ("magnetic_field_z", "gauss"),
        ("radiation_acceleration_x", "cm / s**2"),
        ("radiation_acceleration_y", "cm / s**2"),
        ("radiation_acceleration_z", "cm / s**2"),
    )

    # This set of fields can be applied to any particle type.
    _known_particle_fields = (
        ("particle_position_x", "cm"),
        ("particle_position_y", "cm"),
        ("particle_position_z", "cm"),
        ("particle_velocity_x", "cm / s"),
        ("particle_velocity_y", "cm / s"),
        ("particle_velocity_z", "cm / s"),
        ("particle_mass", "g"),
        ("particle_index", ""),
    )



"""
Tests for frb filters

"""

import numpy as np

import yt
from yt.testing import fake_amr_ds, requires_module


@requires_module("scipy")
def test_white_noise_filter():
    ds = fake_amr_ds(fields=("density",), units=("g/cm**3",))
    p = ds.proj(("gas", "density"), "z")
    frb = p.to_frb((1, "unitary"), 64)
    frb.apply_white_noise()
    frb.apply_white_noise(1e-3)
    frb.render(("gas", "density"))


@requires_module("scipy")
def test_gauss_beam_filter():
    ds = fake_amr_ds(fields=("density",), units=("g/cm**3",))
    p = ds.proj(("gas", "density"), "z")
    frb = p.to_frb((1, "unitary"), 64)
    frb.apply_gauss_beam(sigma=1.0)
    frb.render(("gas", "density"))


@requires_module("scipy")
def test_filter_wiring():

    ds = fake_amr_ds(fields=[("gas", "density")], units=["g/cm**3"])
    p = yt.SlicePlot(ds, "x", "density")

    # Note: frb is a FixedResolutionBuffer object
    frb1 = p.frb
    data_orig = frb1["density"].value

    sigma = 2
    nbeam = 30
    p.frb.apply_gauss_beam(nbeam=nbeam, sigma=sigma)
    frb2 = p.frb
    data_gauss = frb2["density"].value

    p.frb.apply_white_noise()
    frb3 = p.frb
    data_white = frb3["density"].value

    # We check the frb objects are different
    assert frb1 is not frb2
    assert frb1 is not frb3
    assert frb2 is not frb3

    # We check the resulting image are different each time
    assert not np.allclose(data_orig, data_gauss)
    assert not np.allclose(data_orig, data_white)
    assert not np.allclose(data_gauss, data_white)



import yt

# Load the dataset.
ds = yt.load("GalaxyClusterMerger/fiducial_1to3_b0.273d_hdf5_plt_cnt_0175")

# Create density-weighted projections of temperature (weighted line integrals)

for normal in "xyz":
    yt.ProjectionPlot(
        ds, normal, ("gas", "temperature"), weight_field=("gas", "density")
    ).save()



import yt

yt.enable_parallelism()

# Enable parallelism in the script (assuming it was called with
# `mpirun -np <n_procs>` )
yt.enable_parallelism()

# By using wildcards such as ? and * with the load command, we can load up a
# Time Series containing all of these datasets simultaneously.
ts = yt.load("enzo_tiny_cosmology/DD????/DD????")

# Calculate and store density extrema for all datasets along with redshift
# in a data dictionary with entries as tuples

# Create an empty dictionary
data = {}

# Iterate through each dataset in the Time Series (using piter allows it
# to happen in parallel automatically across available processors)
for ds in ts.piter():
    ad = ds.all_data()
    extrema = ad.quantities.extrema(("gas", "density"))

    # Fill the dictionary with extrema and redshift information for each dataset
    data[ds.basename] = (extrema, ds.current_redshift)

# Sort dict by keys
data = {k: v for k, v in sorted(data.items())}

# Print out all the values we calculated.
print("Dataset      Redshift        Density Min      Density Max")
print("---------------------------------------------------------")
for key, val in data.items():
    print(
        "%s       %05.3f          %5.3g g/cm^3   %5.3g g/cm^3"
        % (key, val[1], val[0][0], val[0][1])
    )



yt_to_art = {
    "Density": "HVAR_GAS_DENSITY",
    "TotalEnergy": "HVAR_GAS_ENERGY",
    "GasEnergy": "HVAR_INTERNAL_ENERGY",
    "Pressure": "HVAR_PRESSURE",
    "XMomentumDensity": "HVAR_MOMENTUM_X",
    "YMomentumDensity": "HVAR_MOMENTUM_Y",
    "ZMomentumDensity": "HVAR_MOMENTUM_Z",
    "Gamma": "HVAR_GAMMA",
    "MetalDensitySNIa": "HVAR_METAL_DENSITY_Ia",
    "MetalDensitySNII": "HVAR_METAL_DENSITY_II",
    "Potential": "VAR_POTENTIAL",
    "PotentialHydro": "VAR_POTENTIAL_HYDRO",
    "particle_position_x": "POSITION_X",
    "particle_position_y": "POSITION_Y",
    "particle_position_z": "POSITION_Z",
    "particle_velocity_x": "VELOCITY_X",
    "particle_velocity_y": "VELOCITY_Y",
    "particle_velocity_z": "VELOCITY_Z",
    "particle_mass": "MASS",
    "particle_index": "PID",
    "particle_species": "SPECIES",
    "creation_time": "BIRTH_TIME",
    "particle_mass_initial": "INITIAL_MASS",
    "particle_metallicity1": "METALLICITY_SNIa",
    "particle_metallicity2": "METALLICITY_SNII",
    "stars": "STAR",
    "nbody": "N-BODY",
}
art_to_yt = dict(zip(yt_to_art.values(), yt_to_art.keys()))


class ARTIOconstants:
    def __init__(self):
        self.yr = 365.25 * 86400
        self.Myr = 1.0e6 * self.yr
        self.Gyr = 1.0e9 * self.yr

        self.pc = 3.0856775813e18
        self.kpc = 1.0e3 * self.pc
        self.Mpc = 1.0e6 * self.pc

        self.kms = 1.0e5

        self.mp = 1.672621637e-24
        self.k = 1.3806504e-16
        self.G = 6.67428e-8
        self.c = 2.99792458e10

        self.eV = 1.602176487e-12
        self.amu = 1.660538782e-24
        self.mH = 1.007825 * self.amu
        self.mHe = 4.002602 * self.amu

        self.Msun = 1.32712440018e26 / self.G
        self.Zsun = 0.0199

        self.Yp = 0.24
        self.wmu = 4.0 / (8.0 - 5.0 * self.Yp)
        self.wmu_e = 1.0 / (1.0 - 0.5 * self.Yp)
        self.XH = 1.0 - self.Yp
        self.XHe = 0.25 * self.Yp
        self.gamma = 5.0 / 3.0

        self.sigmaT = 6.6524e-25



"""
AdaptaHOP-specific fields




"""
from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer

m_units = "1e11 * Msun"
r_units = "Mpc"
v_units = "km / s"
l_units = "1e11 * Msun * Mpc * km / s"
e_units = "1e11 * Msun * km**2 / s**2"
dens_units = "1e11 * Msun / Mpc**3"


class AdaptaHOPFieldInfo(FieldInfoContainer):
    known_particle_fields: KnownFieldsT = (
        ("particle_identifier", ("", [], "Halo Identity")),
        ("raw_position_x", (r_units, [], None)),
        ("raw_position_y", (r_units, [], None)),
        ("raw_position_z", (r_units, [], None)),
        ("r", (r_units, [], "Halo radius")),
        ("a", (r_units, [], "Halo semi-major axis")),
        ("b", (r_units, [], "Halo semi-medium axis")),
        ("c", (r_units, [], "Halo semi-minor axis")),
        ("particle_velocity_x", (v_units, [], "Halo velocity x")),
        ("particle_velocity_y", (v_units, [], "Halo velocity y")),
        ("particle_velocity_z", (v_units, [], "Halo velocity z")),
        ("particle_angular_momentum_x", (l_units, [], "Halo Angular Momentum x")),
        ("particle_angular_momentum_y", (l_units, [], "Halo Angular Momentum y")),
        ("particle_angular_momentum_z", (l_units, [], "Halo Angular Momentum z")),
        ("particle_mass", (m_units, [], "Halo mass")),
        ("ek", (e_units, [], "Halo Kinetic Energy")),
        ("ep", (e_units, [], "Halo Gravitational Energy")),
        ("ek", (e_units, [], "Halo Total Energy")),
        ("spin", ("", [], "Halo Spin Parameter")),
        # Virial parameters
        ("virial_radius", (r_units, [], "Halo Virial Radius")),
        ("virial_mass", (m_units, [], "Halo Virial Mass")),
        ("virial_temperature", ("K", [], "Halo Virial Temperature")),
        ("virial_velocity", (v_units, [], "Halo Virial Velocity")),
        # NFW parameters
        ("rho0", (dens_units, [], "Halo NFW Density")),
        ("R_c", (r_units, [], "Halo NFW Scale Radius")),
        ("velocity_dispersion", ("km/s", [], "Velocity Dispersion")),
        ("radius_200", (r_units, [], r"$R_\mathrm{200}$")),
        ("radius_50", (r_units, [], r"$R_\mathrm{50}$")),
        ("mass_200", (m_units, [], r"$M_\mathrm{200}$")),
        ("mass_50", (m_units, [], r"$M_\mathrm{50}$")),
        # Contamination
        ("contaminated", ("", [], "Contaminated")),
        ("m_contam", (m_units, [], "Contaminated Mass")),
    )

    def setup_particle_fields(self, ptype):
        super().setup_particle_fields(ptype)

        # Add particle position
        def generate_pos_field(d):
            shift = self.ds.domain_width[0] / 2

            def closure(field, data):
                return data["halos", f"raw_position_{d}"] + shift

            return closure

        for k in "xyz":
            fun = generate_pos_field(k)
            self.add_field(
                ("halos", f"particle_position_{k}"),
                sampling_type="particle",
                function=fun,
                units="Mpc",
            )



import time

import h5py
import numpy as np

import yt
from yt.utilities.parallel_tools.parallel_analysis_interface import communication_system


@yt.derived_field(
    name="IonizedHydrogen", units="", display_name=r"\frac{\rho_{HII}}{\rho_H}"
)
def IonizedHydrogen(field, data):
    return data[("gas", "HII_Density")] / (
        data[("gas", "HI_Density")] + data[("gas", "HII_Density")]
    )


ts = yt.DatasetSeries("SED800/DD*/*.index", parallel=8)

ionized_z = np.zeros(ts[0].domain_dimensions, dtype="float32")

t1 = time.time()
for ds in ts.piter():
    z = ds.current_redshift
    for g in yt.parallel_objects(ds.index.grids, njobs=16):
        i1, j1, k1 = g.get_global_startindex()  # Index into our domain
        i2, j2, k2 = g.get_global_startindex() + g.ActiveDimensions
        # Look for the newly ionized gas
        newly_ion = (g["IonizedHydrogen"] > 0.999) & (
            ionized_z[i1:i2, j1:j2, k1:k2] < z
        )
        ionized_z[i1:i2, j1:j2, k1:k2][newly_ion] = z
        g.clear_data()

print(f"Iteration completed  {time.time() - t1:0.3e}")
comm = communication_system.communicators[-1]
for i in range(ionized_z.shape[0]):
    ionized_z[i, :, :] = comm.mpi_allreduce(ionized_z[i, :, :], op="max")
    print("Slab % 3i has minimum z of %0.3e" % (i, ionized_z[i, :, :].max()))
t2 = time.time()
print(f"Completed.  {t2 - t1:0.3e}")

if comm.rank == 0:
    f = h5py.File("IonizationCube.h5", mode="w")
    f.create_dataset("/z", data=ionized_z)



import os
from functools import wraps

import numpy as np

from yt.config import ytcfg
from yt.loaders import load
from yt.testing import assert_allclose
from yt.utilities.answer_testing.framework import (
    AnswerTestingTest,
    FieldValuesTest,
    GridValuesTest,
    ProjectionValuesTest,
    can_run_ds,
    temp_cwd,
)


class AssertWrapper:
    """
    Used to wrap a numpy testing assertion, in order to provide a useful name
    for a given assertion test.
    """

    def __init__(self, description, *args):
        # The key here is to add a description attribute, which nose will pick
        # up.
        self.args = args
        self.description = description

    def __call__(self):
        self.args[0](*self.args[1:])


def requires_outputlog(path=".", prefix=""):
    from nose import SkipTest

    def ffalse(func):
        @wraps(func)
        def fskip(*args, **kwargs):
            raise SkipTest

        return fskip

    def ftrue(func):
        @wraps(func)
        def fyielder(*args, **kwargs):
            with temp_cwd(path):
                for t in func(*args, **kwargs):
                    if isinstance(t, AnswerTestingTest):
                        t.prefix = prefix
                    yield t

        return fyielder

    if os.path.exists("OutputLog"):
        return ftrue
    with temp_cwd(path):
        if os.path.exists("OutputLog"):
            return ftrue
    return ffalse


def standard_small_simulation(ds_fn, fields):
    if not can_run_ds(ds_fn):
        return
    dso = [None]
    tolerance = ytcfg.get("yt", "answer_testing_tolerance")
    bitwise = ytcfg.get("yt", "answer_testing_bitwise")
    for field in fields:
        if bitwise:
            yield GridValuesTest(ds_fn, field)
        if "particle" in field:
            continue
        for dobj_name in dso:
            for axis in [0, 1, 2]:
                for weight_field in [None, ("gas", "density")]:
                    yield ProjectionValuesTest(
                        ds_fn, axis, field, weight_field, dobj_name, decimals=tolerance
                    )
            yield FieldValuesTest(ds_fn, field, dobj_name, decimals=tolerance)


class ShockTubeTest:
    def __init__(
        self, data_file, solution_file, fields, left_edges, right_edges, rtol, atol
    ):
        self.solution_file = solution_file
        self.data_file = data_file
        self.fields = fields
        self.left_edges = left_edges
        self.right_edges = right_edges
        self.rtol = rtol
        self.atol = atol

    def __call__(self):
        # Read in the ds
        ds = load(self.data_file)
        exact = self.get_analytical_solution()

        ad = ds.all_data()
        position = ad[("index", "x")]
        for k in self.fields:
            field = ad[k].d
            for xmin, xmax in zip(self.left_edges, self.right_edges):
                mask = (position >= xmin) * (position <= xmax)
                exact_field = np.interp(position[mask], exact["pos"], exact[k])
                myname = f"ShockTubeTest_{k}"
                # yield test vs analytical solution
                yield AssertWrapper(
                    myname,
                    assert_allclose,
                    field[mask],
                    exact_field,
                    self.rtol,
                    self.atol,
                )

    def get_analytical_solution(self):
        # Reads in from file
        pos, dens, vel, pres, inte = np.loadtxt(self.solution_file, unpack=True)
        exact = {}
        exact["pos"] = pos
        exact[("gas", "density")] = dens
        exact[("gas", "velocity_x")] = vel
        exact[("gas", "pressure")] = pres
        exact[("gas", "specific_thermal_energy")] = inte
        return exact



import pytest

from yt.frontends.gdf.api import GDFDataset
from yt.testing import requires_file, units_override_check
from yt.utilities.answer_testing.answer_tests import small_patch_amr

# Test data
sedov = "sedov/sedov_tst_0004.h5"


@pytest.mark.answer_test
class TestGDF:
    answer_file = None
    saved_hashes = None
    answer_version = "000"

    @pytest.mark.usefixtures("hashing")
    @pytest.mark.parametrize("ds", [sedov], indirect=True)
    def test_sedov_tunnel(self, axis, dobj, weight, field, ds):
        self.hashes.update(small_patch_amr(ds, field, weight, axis, dobj))

    @pytest.mark.parametrize("ds", [sedov], indirect=True)
    def test_GDFDataset(self, ds):
        assert isinstance(ds, GDFDataset)

    @requires_file(sedov)
    def test_units_override(self):
        units_override_check(sedov)



import re

import numpy as np
from more_itertools import always_iterable

from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer,
    YTSelectionContainer3D,
)
from yt.data_objects.static_output import Dataset
from yt.funcs import iter_fields, validate_object, validate_sequence
from yt.geometry.selection_routines import points_in_cells
from yt.utilities.exceptions import YTIllDefinedCutRegion
from yt.utilities.on_demand_imports import _scipy


class YTCutRegion(YTSelectionContainer3D):
    """
    This is a data object designed to allow individuals to apply logical
    operations to fields and filter as a result of those cuts.

    Parameters
    ----------
    data_source : YTSelectionContainer3D
        The object to which cuts will be applied.
    conditionals : list of strings
        A list of conditionals that will be evaluated.  In the namespace
        available, these conditionals will have access to 'obj' which is a data
        object of unknown shape, and they must generate a boolean array.  For
        instance, conditionals = ["obj[('gas', 'temperature')] < 1e3"]

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> sp = ds.sphere("max", (1.0, "Mpc"))
    >>> cr = ds.cut_region(sp, ["obj[('gas', 'temperature')] < 1e3"])
    """

    _type_name = "cut_region"
    _con_args = ("base_object", "conditionals")
    _derived_quantity_chunking = "all"

    def __init__(
        self,
        data_source,
        conditionals,
        ds=None,
        field_parameters=None,
        base_object=None,
        locals=None,
    ):
        if locals is None:
            locals = {}
        validate_object(data_source, YTSelectionContainer)
        validate_sequence(conditionals)
        for condition in conditionals:
            validate_object(condition, str)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(base_object, YTSelectionContainer)

        self.conditionals = list(always_iterable(conditionals))
        if isinstance(data_source, YTCutRegion):
            # If the source is also a cut region, add its conditionals
            # and set the source to be its source.
            # Preserve order of conditionals.
            self.conditionals = data_source.conditionals + self.conditionals
            data_source = data_source.base_object

        super().__init__(
            data_source.center, ds, field_parameters, data_source=data_source
        )
        self.filter_fields = self._check_filter_fields()
        self.base_object = data_source
        self.locals = locals
        self._selector = None
        # Need to interpose for __getitem__, fwidth, fcoords, icoords, iwidth,
        # ires and get_data

    def _check_filter_fields(self):
        fields = []
        for cond in self.conditionals:
            for field in re.findall(r"\[([A-Za-z0-9_,.'\"\(\)]+)\]", cond):
                fd = field.replace('"', "").replace("'", "")
                if "," in fd:
                    fd = tuple(fd.strip("()").split(","))
                fd = self.ds._get_field_info(fd)
                if fd.sampling_type == "particle" or fd.is_sph_field:
                    raise RuntimeError(
                        f"cut_region requires a mesh-based field, "
                        f"but {fd.name} is a particle field! Use "
                        f"a particle filter instead. "
                    )
                fields.append(fd.name)
        return fields

    def chunks(self, fields, chunking_style, **kwargs):
        # We actually want to chunk the sub-chunk, not ourselves.  We have no
        # chunks to speak of, as we do not data IO.
        for chunk in self.index._chunk(self.base_object, chunking_style, **kwargs):
            with self.base_object._chunked_read(chunk):
                with self._chunked_read(chunk):
                    self.get_data(fields)
                    yield self

    def get_data(self, fields=None):
        fields = list(iter_fields(fields))
        self.base_object.get_data(fields)
        ind = self._cond_ind
        for field in fields:
            f = self.base_object[field]
            if f.shape != ind.shape:
                parent = getattr(self, "parent", self.base_object)
                self.field_data[field] = parent[field][self._part_ind(field[0])]
            else:
                self.field_data[field] = self.base_object[field][ind]

    @property
    def blocks(self):
        # We have to take a slightly different approach here.  Note that all
        # that .blocks has to yield is a 3D array and a mask.
        for obj, m in self.base_object.blocks:
            m = m.copy()
            with obj._field_parameter_state(self.field_parameters):
                for cond in self.conditionals:
                    ss = eval(cond)
                    m = np.logical_and(m, ss, m)
            if not np.any(m):
                continue
            yield obj, m

    @property
    def _cond_ind(self):
        ind = None
        obj = self.base_object
        locals = self.locals.copy()
        if "obj" in locals:
            raise RuntimeError(
                '"obj" has been defined in the "locals" ; '
                "this is not supported, please rename the variable."
            )
        locals["obj"] = obj
        with obj._field_parameter_state(self.field_parameters):
            for cond in self.conditionals:
                res = eval(cond, locals)
                if ind is None:
                    ind = res
                if ind.shape != res.shape:
                    raise YTIllDefinedCutRegion(self.conditionals)
                np.logical_and(res, ind, ind)
        return ind

    def _part_ind_KDTree(self, ptype):
        """Find the particles in cells using a KDTree approach."""
        parent = getattr(self, "parent", self.base_object)
        units = "code_length"

        pos = np.stack(
            [
                self[("index", "x")].to(units),
                self[("index", "y")].to(units),
                self[("index", "z")].to(units),
            ],
            axis=1,
        ).value
        dx = np.stack(
            [
                self[("index", "dx")].to(units),
                self[("index", "dy")].to(units),
                self[("index", "dz")].to(units),
            ],
            axis=1,
        ).value
        ppos = np.stack(
            [
                parent[(ptype, "particle_position_x")],
                parent[(ptype, "particle_position_y")],
                parent[(ptype, "particle_position_z")],
            ],
            axis=1,
        ).value

        mask = np.zeros(ppos.shape[0], dtype=bool)
        levels = self[("index", "grid_level")].astype("int32").value
        if levels.size == 0:
            return mask

        levelmin = levels.min()
        levelmax = levels.max()

        for lvl in range(levelmax, levelmin - 1, -1):
            # Filter out cells not in the current level
            lvl_mask = levels == lvl
            dx_loc = dx[lvl_mask]
            pos_loc = pos[lvl_mask]

            grid_tree = _scipy.spatial.cKDTree(pos_loc, boxsize=1)

            # Compute closest cell for all remaining particles
            dist, icell = grid_tree.query(
                ppos[~mask], distance_upper_bound=dx_loc.max(), p=np.inf
            )
            mask_loc = np.isfinite(dist[:])

            # Check that particles within dx of a cell are in it
            i = icell[mask_loc]
            dist = np.abs(ppos[~mask][mask_loc, :] - pos_loc[i])
            tmp_mask = np.all(dist <= (dx_loc[i] / 2), axis=1)

            mask_loc[mask_loc] = tmp_mask

            # Update the particle mask with particles found at this level
            mask[~mask] |= mask_loc

        return mask

    def _part_ind_brute_force(self, ptype):
        parent = getattr(self, "parent", self.base_object)
        units = "code_length"
        mask = points_in_cells(
            self[("index", "x")].to(units),
            self[("index", "y")].to(units),
            self[("index", "z")].to(units),
            self[("index", "dx")].to(units),
            self[("index", "dy")].to(units),
            self[("index", "dz")].to(units),
            parent[(ptype, "particle_position_x")].to(units),
            parent[(ptype, "particle_position_y")].to(units),
            parent[(ptype, "particle_position_z")].to(units),
        )

        return mask

    def _part_ind(self, ptype):
        # If scipy is installed, use the fast KD tree
        # implementation. Else, fall back onto the direct
        # brute-force algorithm.
        try:
            _scipy.spatial.KDTree
            return self._part_ind_KDTree(ptype)
        except ImportError:
            return self._part_ind_brute_force(ptype)

    @property
    def icoords(self):
        return self.base_object.icoords[self._cond_ind, :]

    @property
    def fcoords(self):
        return self.base_object.fcoords[self._cond_ind, :]

    @property
    def ires(self):
        return self.base_object.ires[self._cond_ind]

    @property
    def fwidth(self):
        return self.base_object.fwidth[self._cond_ind, :]

    def _get_bbox(self):
        """
        Get the bounding box for the cut region. Here we just use
        the bounding box for the source region.
        """
        return self.base_object._get_bbox()



import numpy as np
from matplotlib import pyplot as plt

import yt
from yt.data_objects.particle_filters import add_particle_filter


def formed_star(pfilter, data):
    filter = data["all", "creation_time"] > 0
    return filter


add_particle_filter(
    "formed_star", function=formed_star, filtered_type="all", requires=["creation_time"]
)

filename = "IsolatedGalaxy/galaxy0030/galaxy0030"

ds = yt.load(filename)
ds.add_particle_filter("formed_star")
ad = ds.all_data()
masses = ad["formed_star", "particle_mass"].in_units("Msun")
formation_time = ad["formed_star", "creation_time"].in_units("yr")

time_range = [0, 5e8]  # years
n_bins = 1000
hist, bins = np.histogram(
    formation_time,
    bins=n_bins,
    range=time_range,
)
inds = np.digitize(formation_time, bins=bins)
time = (bins[:-1] + bins[1:]) / 2

sfr = np.array(
    [masses[inds == j + 1].sum() / (bins[j + 1] - bins[j]) for j in range(len(time))]
)
sfr[sfr == 0] = np.nan

plt.plot(time / 1e6, sfr)
plt.xlabel("Time  [Myr]")
plt.ylabel(r"SFR  [M$_\odot$ yr$^{-1}$]")
plt.savefig("filter_sfr.png")



import numpy as np

from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer,
    YTSelectionContainer3D,
)
from yt.data_objects.static_output import Dataset
from yt.funcs import validate_center, validate_object, validate_sequence


class YTDataCollection(YTSelectionContainer3D):
    """
    By selecting an arbitrary *object_list*, we can act on those grids.
    Child cells are not returned.
    """

    _type_name = "data_collection"
    _con_args = ("_obj_list",)

    def __init__(
        self, obj_list, ds=None, field_parameters=None, data_source=None, center=None
    ):
        validate_sequence(obj_list)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        if center is not None:
            validate_center(center)
        YTSelectionContainer3D.__init__(self, center, ds, field_parameters, data_source)
        self._obj_ids = np.array([o.id - o._id_offset for o in obj_list], dtype="int64")
        self._obj_list = obj_list



import os
import sys
import tarfile
import time

import pytest

from yt.config import ytcfg
from yt.loaders import load_archive
from yt.sample_data.api import _download_sample_data_file, get_data_registry_table
from yt.testing import requires_module_pytest
from yt.utilities.exceptions import YTUnidentifiedDataType


@pytest.fixture()
def data_registry():
    yield get_data_registry_table()


@pytest.fixture()
def tmp_data_dir(tmp_path):
    pre_test_data_dir = ytcfg["yt", "test_data_dir"]
    ytcfg.set("yt", "test_data_dir", str(tmp_path))

    yield tmp_path

    ytcfg.set("yt", "test_data_dir", pre_test_data_dir)


# Note: ratarmount cannot currently be installed on Windows as of v0.8.1
@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="ratarmount cannot currently be installed on Windows as of v0.8.1",
)
@pytest.mark.skipif(
    os.environ.get("JENKINS_HOME") is not None,
    reason="Archive mounting times out on Jenkins.",
)
@requires_module_pytest("pooch", "ratarmount")
@pytest.mark.parametrize(
    "fn, exact_loc, class_",
    [
        (
            "ToroShockTube.tar.gz",
            "ToroShockTube/DD0001/data0001",
            "EnzoDataset",
        ),
        (
            "ramses_sink_00016.tar.gz",
            "ramses_sink_00016/output_00016",
            "RAMSESDataset",
        ),
    ],
)
@pytest.mark.parametrize("archive_suffix", ["", ".gz"])
def test_load_archive(
    fn, exact_loc, class_: str, archive_suffix, tmp_data_dir, data_registry
):
    # Download the sample .tar.gz'd file
    targz_path = _download_sample_data_file(filename=fn)
    tar_path = targz_path.with_suffix(archive_suffix)

    if tar_path != targz_path:
        # Open the tarfile and uncompress it to .tar, .tar.gz, and .tar.bz2 files
        with tarfile.open(targz_path, mode="r:*") as targz:
            mode = "w" + archive_suffix.replace(".", ":")
            with tarfile.open(tar_path, mode=mode) as tar:
                for member in targz.getmembers():
                    content = targz.extractfile(member)
                    tar.addfile(member, fileobj=content)

    # Now try to open the .tar.* files
    warn_msg = "The 'load_archive' function is still experimental and may be unstable."
    with pytest.warns(UserWarning, match=warn_msg):
        ds = load_archive(tar_path, exact_loc, mount_timeout=10)
    assert type(ds).__name__ == class_

    # Make sure the index is readable
    ds.index

    # Check cleanup
    mount_path = tar_path.with_name(tar_path.name + ".mount")
    assert mount_path.is_mount()

    ## Manually dismount
    ds.dismount()

    ## The dismounting happens concurrently, wait a few sec.
    time.sleep(2)

    ## Mount path should not exist anymore *and* have been deleted
    assert not mount_path.is_mount()
    assert not mount_path.exists()


@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="ratarmount cannot currently be installed on Windows as of v0.8.1",
)
@pytest.mark.skipif(
    os.environ.get("JENKINS_HOME") is not None,
    reason="Archive mounting times out on Jenkins.",
)
@pytest.mark.filterwarnings(
    "ignore:The 'load_archive' function is still experimental and may be unstable."
)
@requires_module_pytest("pooch", "ratarmount")
def test_load_invalid_archive(tmp_data_dir, data_registry):
    # Archive does not exist
    with pytest.raises(FileNotFoundError):
        load_archive("this_file_does_not_exist.tar.gz", "invalid_location")

    targz_path = _download_sample_data_file(filename="ToroShockTube.tar.gz")
    # File does not exist
    with pytest.raises(FileNotFoundError):
        load_archive(targz_path, "invalid_location")

    # File exists but is not recognized
    with pytest.raises(YTUnidentifiedDataType):
        load_archive(targz_path, "ToroShockTube/DD0001/data0001.memorymap")



from typing import Dict

import numpy as np

from yt.geometry.selection_routines import GridSelector
from yt.utilities.io_handler import BaseIOHandler
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _h5py as h5py

_convert_mass = ("particle_mass", "mass")

_particle_position_names: Dict[str, str] = {}


class IOHandlerPackedHDF5(BaseIOHandler):

    _dataset_type = "enzo_packed_3d"
    _base = slice(None)
    _field_dtype = "float64"

    def _read_field_names(self, grid):
        if grid.filename is None:
            return []
        f = h5py.File(grid.filename, mode="r")
        try:
            group = f["/Grid%08i" % grid.id]
        except KeyError:
            group = f
        fields = []
        dtypes = set()
        add_io = "io" in grid.ds.particle_types
        add_dm = "DarkMatter" in grid.ds.particle_types
        for name, v in group.items():
            # NOTE: This won't work with 1D datasets or references.
            # For all versions of Enzo I know about, we can assume all floats
            # are of the same size.  So, let's grab one.
            if not hasattr(v, "shape") or v.dtype == "O":
                continue
            elif len(v.dims) == 1:
                if grid.ds.dimensionality == 1:
                    fields.append(("enzo", str(name)))
                elif add_io:
                    fields.append(("io", str(name)))
                elif add_dm:
                    fields.append(("DarkMatter", str(name)))
            else:
                fields.append(("enzo", str(name)))
                dtypes.add(v.dtype)

        if len(dtypes) == 1:
            # Now, if everything we saw was the same dtype, we can go ahead and
            # set it here.  We do this because it is a HUGE savings for 32 bit
            # floats, since our numpy copying/casting is way faster than
            # h5py's, for some reason I don't understand.  This does *not* need
            # to be correct -- it will get fixed later -- it just needs to be
            # okay for now.
            self._field_dtype = list(dtypes)[0]
        f.close()
        return fields

    @property
    def _read_exception(self):
        return (KeyError,)

    def _read_particle_coords(self, chunks, ptf):
        yield from (
            (ptype, xyz, 0.0)
            for ptype, xyz in self._read_particle_fields(chunks, ptf, None)
        )

    def _read_particle_fields(self, chunks, ptf, selector):
        chunks = list(chunks)
        for chunk in chunks:  # These should be organized by grid filename
            f = None
            for g in chunk.objs:
                if g.filename is None:
                    continue
                if f is None:
                    # print("Opening (read) %s" % g.filename)
                    f = h5py.File(g.filename, mode="r")
                nap = sum(g.NumberOfActiveParticles.values())
                if g.NumberOfParticles == 0 and nap == 0:
                    continue
                ds = f.get("/Grid%08i" % g.id)
                for ptype, field_list in sorted(ptf.items()):
                    if ptype == "io":
                        if g.NumberOfParticles == 0:
                            continue
                        pds = ds
                    elif ptype == "DarkMatter":
                        if g.NumberOfActiveParticles[ptype] == 0:
                            continue
                        pds = ds
                    elif not g.NumberOfActiveParticles[ptype]:
                        continue
                    else:
                        for pname in ["Active Particles", "Particles"]:
                            pds = ds.get(f"{pname}/{ptype}")
                            if pds is not None:
                                break
                        else:
                            raise RuntimeError(
                                "Could not find active particle group in data."
                            )
                    pn = _particle_position_names.get(ptype, r"particle_position_%s")
                    x, y, z = (
                        np.asarray(pds.get(pn % ax)[()], dtype="=f8") for ax in "xyz"
                    )
                    if selector is None:
                        # This only ever happens if the call is made from
                        # _read_particle_coords.
                        yield ptype, (x, y, z)
                        continue
                    mask = selector.select_points(x, y, z, 0.0)
                    if mask is None:
                        continue
                    for field in field_list:
                        data = np.asarray(pds.get(field)[()], "=f8")
                        if field in _convert_mass:
                            data *= g.dds.prod(dtype="f8")
                        yield (ptype, field), data[mask]
            if f:
                f.close()

    def io_iter(self, chunks, fields):
        h5_dtype = self._field_dtype
        for chunk in chunks:
            fid = None
            filename = -1
            for obj in chunk.objs:
                if obj.filename is None:
                    continue
                if obj.filename != filename:
                    # Note one really important thing here: even if we do
                    # implement LRU caching in the _read_obj_field function,
                    # we'll still be doing file opening and whatnot.  This is a
                    # problem, but one we can return to.
                    if fid is not None:
                        fid.close()
                    fid = h5py.h5f.open(
                        obj.filename.encode("latin-1"), h5py.h5f.ACC_RDONLY
                    )
                    filename = obj.filename
                for field in fields:
                    nodal_flag = self.ds.field_info[field].nodal_flag
                    dims = obj.ActiveDimensions[::-1] + nodal_flag[::-1]
                    data = np.empty(dims, dtype=h5_dtype)
                    yield field, obj, self._read_obj_field(obj, field, (fid, data))
        if fid is not None:
            fid.close()

    def _read_obj_field(self, obj, field, fid_data):
        if fid_data is None:
            fid_data = (None, None)
        fid, data = fid_data
        if fid is None:
            close = True
            fid = h5py.h5f.open(obj.filename.encode("latin-1"), h5py.h5f.ACC_RDONLY)
        else:
            close = False
        if data is None:
            data = np.empty(obj.ActiveDimensions[::-1], dtype=self._field_dtype)
        ftype, fname = field
        try:
            node = "/Grid%08i/%s" % (obj.id, fname)
            dg = h5py.h5d.open(fid, node.encode("latin-1"))
        except KeyError:
            if fname == "Dark_Matter_Density":
                data[:] = 0
                return data.T
            raise
        dg.read(h5py.h5s.ALL, h5py.h5s.ALL, data)
        # I don't know why, but on some installations of h5py this works, but
        # on others, nope.  Doesn't seem to be a version thing.
        # dg.close()
        if close:
            fid.close()
        return data.T


class IOHandlerPackedHDF5GhostZones(IOHandlerPackedHDF5):
    _dataset_type = "enzo_packed_3d_gz"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        NGZ = self.ds.parameters.get("NumberOfGhostZones", 3)
        self._base = (slice(NGZ, -NGZ), slice(NGZ, -NGZ), slice(NGZ, -NGZ))

    def _read_obj_field(self, *args, **kwargs):
        return super()._read_obj_field(*args, **kwargs)[self._base]


class IOHandlerInMemory(BaseIOHandler):

    _dataset_type = "enzo_inline"

    def __init__(self, ds, ghost_zones=3):
        self.ds = ds
        import enzo

        self.enzo = enzo
        self.grids_in_memory = enzo.grid_data
        self.old_grids_in_memory = enzo.old_grid_data
        self.my_slice = (
            slice(ghost_zones, -ghost_zones),
            slice(ghost_zones, -ghost_zones),
            slice(ghost_zones, -ghost_zones),
        )
        BaseIOHandler.__init__(self, ds)

    def _read_field_names(self, grid):
        fields = []
        add_io = "io" in grid.ds.particle_types
        for name, v in self.grids_in_memory[grid.id].items():
            # NOTE: This won't work with 1D datasets or references.
            if not hasattr(v, "shape") or v.dtype == "O":
                continue
            elif v.ndim == 1:
                if grid.ds.dimensionality == 1:
                    fields.append(("enzo", str(name)))
                elif add_io:
                    fields.append(("io", str(name)))
            else:
                fields.append(("enzo", str(name)))
        return fields

    def _read_fluid_selection(self, chunks, selector, fields, size):
        rv = {}
        # Now we have to do something unpleasant
        chunks = list(chunks)
        if isinstance(selector, GridSelector):
            if not (len(chunks) == len(chunks[0].objs) == 1):
                raise RuntimeError
            g = chunks[0].objs[0]
            for ftype, fname in fields:
                rv[(ftype, fname)] = self.grids_in_memory[g.id][fname].swapaxes(0, 2)
            return rv
        if size is None:
            size = sum(g.count(selector) for chunk in chunks for g in chunk.objs)
        for field in fields:
            ftype, fname = field
            fsize = size
            rv[field] = np.empty(fsize, dtype="float64")
        ng = sum(len(c.objs) for c in chunks)
        mylog.debug(
            "Reading %s cells of %s fields in %s grids",
            size,
            [f2 for f1, f2 in fields],
            ng,
        )

        ind = 0
        for chunk in chunks:
            for g in chunk.objs:
                # We want a *hard error* here.
                # if g.id not in self.grids_in_memory: continue
                for field in fields:
                    ftype, fname = field
                    data_view = self.grids_in_memory[g.id][fname][
                        self.my_slice
                    ].swapaxes(0, 2)
                    nd = g.select(selector, data_view, rv[field], ind)
                ind += nd
        assert ind == fsize
        return rv

    def _read_particle_coords(self, chunks, ptf):
        chunks = list(chunks)
        for chunk in chunks:  # These should be organized by grid filename
            for g in chunk.objs:
                if g.id not in self.grids_in_memory:
                    continue
                nap = sum(g.NumberOfActiveParticles.values())
                if g.NumberOfParticles == 0 and nap == 0:
                    continue
                for ptype in sorted(ptf):
                    x, y, z = (
                        self.grids_in_memory[g.id]["particle_position_x"],
                        self.grids_in_memory[g.id]["particle_position_y"],
                        self.grids_in_memory[g.id]["particle_position_z"],
                    )
                    yield ptype, (x, y, z), 0.0

    def _read_particle_fields(self, chunks, ptf, selector):
        chunks = list(chunks)
        for chunk in chunks:  # These should be organized by grid filename
            for g in chunk.objs:
                if g.id not in self.grids_in_memory:
                    continue
                nap = sum(g.NumberOfActiveParticles.values())
                if g.NumberOfParticles == 0 and nap == 0:
                    continue
                for ptype, field_list in sorted(ptf.items()):
                    x, y, z = (
                        self.grids_in_memory[g.id]["particle_position_x"],
                        self.grids_in_memory[g.id]["particle_position_y"],
                        self.grids_in_memory[g.id]["particle_position_z"],
                    )
                    mask = selector.select_points(x, y, z, 0.0)
                    if mask is None:
                        continue
                    for field in field_list:
                        data = self.grids_in_memory[g.id][field]
                        if field in _convert_mass:
                            data = data * g.dds.prod(dtype="f8")
                        yield (ptype, field), data[mask]


class IOHandlerPacked2D(IOHandlerPackedHDF5):

    _dataset_type = "enzo_packed_2d"
    _particle_reader = False

    def _read_data_set(self, grid, field):
        f = h5py.File(grid.filename, mode="r")
        ds = f["/Grid%08i/%s" % (grid.id, field)][:]
        f.close()
        return ds.transpose()[:, :, None]

    def _read_fluid_selection(self, chunks, selector, fields, size):
        rv = {}
        # Now we have to do something unpleasant
        chunks = list(chunks)
        if isinstance(selector, GridSelector):
            if not (len(chunks) == len(chunks[0].objs) == 1):
                raise RuntimeError
            g = chunks[0].objs[0]
            f = h5py.File(g.filename, mode="r")
            gds = f.get("/Grid%08i" % g.id)
            for ftype, fname in fields:
                rv[(ftype, fname)] = np.atleast_3d(gds.get(fname)[()].transpose())
            f.close()
            return rv
        if size is None:
            size = sum(g.count(selector) for chunk in chunks for g in chunk.objs)
        for field in fields:
            ftype, fname = field
            fsize = size
            rv[field] = np.empty(fsize, dtype="float64")
        ng = sum(len(c.objs) for c in chunks)
        mylog.debug(
            "Reading %s cells of %s fields in %s grids",
            size,
            [f2 for f1, f2 in fields],
            ng,
        )
        ind = 0
        for chunk in chunks:
            f = None
            for g in chunk.objs:
                if f is None:
                    # print("Opening (count) %s" % g.filename)
                    f = h5py.File(g.filename, mode="r")
                gds = f.get("/Grid%08i" % g.id)
                if gds is None:
                    gds = f
                for field in fields:
                    ftype, fname = field
                    ds = np.atleast_3d(gds.get(fname)[()].transpose())
                    nd = g.select(selector, ds, rv[field], ind)  # caches
                ind += nd
            f.close()
        return rv


class IOHandlerPacked1D(IOHandlerPackedHDF5):

    _dataset_type = "enzo_packed_1d"
    _particle_reader = False

    def _read_data_set(self, grid, field):
        f = h5py.File(grid.filename, mode="r")
        ds = f["/Grid%08i/%s" % (grid.id, field)][:]
        f.close()
        return ds.transpose()[:, None, None]



import yt

# Load the dataset
ds = yt.load("GasSloshing/sloshing_nomag2_hdf5_plt_cnt_0150")

# Create gas density slices of several fields along the x axis simultaneously
yt.SlicePlot(
    ds,
    "x",
    [("gas", "density"), ("gas", "temperature"), ("gas", "pressure")],
    width=(800.0, "kpc"),
).save()
